"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from re import compile
from re import match as re_match
from typing import Any
from typing import Optional
from typing import Union

from .empty import Empty
from .types import DictStrAny
from .types import LDictStrAny



_INTEGER = compile(r'^\d+$')
_INDICES = (list, tuple, dict)
_RECURSE = dict



_SETABLE = Union[
    DictStrAny,
    list[Any]]

_GETABLE = Union[
    tuple[Any, ...],
    _SETABLE]



def getate(
    source: _GETABLE,
    path: str,
    default: Optional[Any] = None,
    delim: str = '/',
) -> Any:  # noqa: ANN401
    """
    Collect the value within the dictionary using notation.

    Example
    -------
    >>> source = {'foo': {'bar': 'baz'}}
    >>> getate(source, 'foo/bar')
    'baz'

    Example
    -------
    >>> source = {'foo': ['bar', 'baz']}
    >>> getate(source, 'foo/1')
    'baz'

    :param source: Dictionary object processed in notation.
    :param path: Path to the value within the source object.
    :param default: Value to use if none is found in source.
    :param delim: Override default delimiter between parts.
    :returns: Value that was located within provided source.
    """

    sourze: Any = source

    split = path.split(delim)

    length = len(split)


    items = enumerate(split)

    for index, base in items:

        if sourze is Empty:
            return default


        indices = isinstance(
            sourze, _INDICES)

        if (indices is False
                and index < length):
            return default


        recurse = isinstance(
            sourze, _RECURSE)

        if recurse is False:

            with suppress(IndexError):
                sourze = sourze[int(base)]
                continue


        if base not in sourze:
            sourze = Empty
            continue

        sourze = sourze[base]


    return (
        default if sourze is Empty
        else sourze)



def setate(
    source: _SETABLE,
    path: str,
    value: Any,  # noqa: ANN401
    delim: str = '/',
) -> None:
    """
    Define the value within the dictionary using notation.

    Example
    -------
    >>> source = {'foo': {'bar': 'baz'}}
    >>> source['foo']['bar']
    'baz'
    >>> setate(source, 'foo/bar', 'bop')
    >>> source['foo']['bar']
    'bop'

    :param source: Dictionary object processed in notation.
    :param path: Path to the value within the source object.
    :param value: Value which will be defined at noted point.
    :param delim: Override default delimiter between parts.
    """

    _setvalue(source, path, value, delim)



def delate(
    source: _SETABLE,
    path: str,
    delim: str = '/',
) -> None:
    """
    Delete the value within the dictionary using notation.

    Example
    -------
    >>> source = {'foo': {'bar': 'baz'}}
    >>> delate(source, 'foo/bar')
    >>> source
    {'foo': {}}

    :param source: Dictionary object processed in notation.
    :param path: Path to the value within the source object.
    :param delim: Override default delimiter between parts.
    """

    split = path.split(delim)

    with suppress(KeyError, IndexError):


        for part in split[:-1]:

            setable = isinstance(
                source, dict | list)

            if setable is False:
                raise ValueError('source')

            if isinstance(source, list):
                source = source[int(part)]

            elif isinstance(source, dict):
                source = source[part]


        part = split[-1]

        setable = isinstance(
            source, dict | list)

        if setable is False:
            raise ValueError('source')

        if isinstance(source, dict):
            del source[part]

        if isinstance(source, list):
            del source[int(part)]



def impate(  # noqa: CFQ001,CFQ004
    source: DictStrAny | LDictStrAny,
    delim: str = '/',
    parent: Optional[str] = None,
    *,
    implode_list: bool = True,
    recurse_list: bool = True,
) -> DictStrAny | LDictStrAny:
    """
    Implode the dictionary into a single depth of notation.

    Example
    -------
    >>> impate({'foo': {'bar': 'baz'}})
    {'foo/bar': 'baz'}

    :param source: Dictionary object processed in notation.
    :param delim: Override default delimiter between parts.
    :param parent: Parent key prefix for downstream update.
    :param implode_list: Determine whether list is imploded.
    :param recurse_list: Determine whether flatten in list.
    :returns: New dictionary that was recursively imploded.
        It is also possible that a list of dictionary will
        be returned when provided and implode_list is False.
    """

    _implode = implode_list
    _recurse = recurse_list


    def _proclist(
        source: list[Any],
        delim: str,
        parent: Optional[str],
    ) -> DictStrAny | list[Any]:


        if _implode is False:

            process = [
                (impate(
                    item, delim,
                    implode_list=_implode,
                    recurse_list=_recurse)
                 if isinstance(item, dict | list)
                 and _recurse is True
                 else item)
                for item in source]

            return (
                {parent: process}
                if parent is not None
                else process)


        returned: DictStrAny = {}

        for i, item in enumerate(source):

            key = (
                f'{parent}{delim}{i}'
                if parent is not None
                else str(i))

            if (isinstance(item, dict | list)
                    and _recurse is True):

                implode = impate(
                    item, delim, key,
                    implode_list=_implode,
                    recurse_list=_recurse)

                assert isinstance(implode, dict)

                returned.update(implode)

            else:
                returned[key] = item

        return returned


    def _procdict(
        source: DictStrAny,
        delim: str,
        parent: Optional[str],
    ) -> DictStrAny:

        returned: DictStrAny = {}

        for key, value in source.items():

            key = (
                f'{parent}{delim}{key}'
                if parent is not None
                else key)

            if isinstance(value, dict):

                implode = impate(
                    value, delim, key,
                    implode_list=_implode,
                    recurse_list=_recurse)

                assert isinstance(implode, dict)

                returned |= implode

            elif isinstance(value, list):

                process = _proclist(
                    value, delim, key)

                returned |= (
                    {key: process}
                    if not isinstance(process, dict)
                    else process)

            else:
                returned[key] = value

        return returned


    if isinstance(source, dict):
        return _procdict(
            source, delim, parent)

    if isinstance(source, list):
        return _proclist(
            source, delim, parent)

    raise ValueError('source')



def expate(
    source: DictStrAny,
    delim: str = '/',
) -> DictStrAny:
    """
    Explode the dictionary from a single depth of notation.

    Example
    -------
    >>> expate({'foo/bar': 'baz'})
    {'foo': {'bar': 'baz'}}

    :param source: Dictionary object processed in notation.
    :param delim: Override default delimiter between parts.
    :returns: New dictionary that was recursively exploded.
    """

    returned: DictStrAny = {}


    items = source.items()

    for key, value in items:

        if isinstance(value, list):
            value = [
                (expate(x, delim)
                 if isinstance(x, dict)
                 else x)
                for x in value]

        if isinstance(value, dict):
            value = expate(value, delim)

        setate(
            returned, key,
            value, delim)


    return returned



def _setpath(
    source: _SETABLE,
    path: str,
    value: Any,  # noqa: ANN401
    delim: str = '/',
) -> None:
    """
    Define the value within the dictionary using notation.

    .. note::
       This is a private helper function that could change.

    :param source: Dictionary object processed in notation.
    :param path: Path to the value within the source object.
    :param value: Value which will be defined at noted point.
    :param delim: Override default delimiter between parts.
    """


    setable = isinstance(
        source, dict | list)

    assert setable is True


    base, path = (
        path.split(delim, 1))

    next = (
        path.split(delim, 1)[0])


    default: _SETABLE = {}

    if re_match(_INTEGER, next):
        default = []

    update: Any


    if isinstance(source, list):

        length = len(source)
        index = int(base)

        update = default

        with suppress(IndexError):
            update = source[index]

        _setvalue(
            update, path,
            value, delim)

        if length == index:
            source.append(update)

        elif length > index:
            source[index] = update


    elif isinstance(source, dict):

        update = default

        with suppress(KeyError):
            update = source[base]

        _setvalue(
            update, path,
            value, delim)

        source[base] = update



def _setvalue(
    source: _SETABLE,
    path: str,
    value: Any,  # noqa: ANN401
    delim: str = '/',
) -> None:
    """
    Define the value within the dictionary using notation.

    .. note::
       This is a private helper function that could change.

    :param source: Dictionary object processed in notation.
    :param path: Path to the value within the source object.
    :param value: Value which will be defined at noted point.
    :param delim: Override default delimiter between parts.
    """


    setable = isinstance(
        source, dict | list)

    if setable is False:
        raise ValueError('source')


    if delim in path:
        return _setpath(
            source, path,
            value, delim)


    if isinstance(source, list):

        length = len(source)
        index = int(path)

        if index > length:
            raise IndexError(index)

        if length == index:
            source.append(value)

        elif length > index:
            source[index] = value


    if isinstance(source, dict):
        source[path] = value



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from pytest import fixture

from . import SAMPLES
from ..files import ConfigFile
from ..files import ConfigFiles
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def files(
    config_path: Path,
) -> ConfigFiles:
    """
    Construct the instance for use in the downstream tests.

    :param config_path: Custom fixture for populating paths.
    :returns: Newly constructed instance of related class.
    """

    return ConfigFiles([
        f'{SAMPLES}/wayne/bwayne.yml',
        f'{SAMPLES}/stark/tstark.yml'])



def test_ConfigFile(
    config_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config_path: Custom fixture for populating paths.
    """

    file = ConfigFile(
        f'{config_path}/config.yml')


    attrs = lattrs(file)

    assert attrs == [
        'path',
        'config']


    assert inrepr(
        'files.ConfigFile object',
        file)

    assert isinstance(
        hash(file), int)

    assert instr(
        'files.ConfigFile object',
        file)


    assert file.path.name == 'config.yml'

    assert len(file.config) == 3



def test_ConfigFiles(
    files: ConfigFiles,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param files: Custom fixture for the configuration files.
    """


    attrs = lattrs(files)

    assert attrs == [
        'paths',
        'config',
        '_ConfigFiles__merge']


    assert inrepr(
        'files.ConfigFiles object',
        files)

    assert isinstance(
        hash(files), int)

    assert instr(
        'files.ConfigFiles object',
        files)


    assert len(files.paths) == 2

    assert len(files.config) == 2

    assert files.merge == {
        'name': 'Bruce Wayne'}



def test_ConfigFiles_cover(
    files: ConfigFiles,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param files: Custom fixture for the configuration files.
    """

    merge1 = files.merge
    merge2 = files.merge

    assert merge1 is not merge2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import Optional

from ..utils import read_text



IMAGES = (
    Path(__file__).parent
    / 'images')

SCRIPTS = (
    Path(__file__).parent
    / 'scripts')

STYLES = (
    Path(__file__).parent
    / 'styles')



class Content:
    """
    Access common content used with application interfaces.
    """


    @classmethod
    def images(
        cls,
        name: str,
    ) -> Optional[str]:
        """
        Return the contents requested stored within the project.

        :param name: Which file in the project will be returned.
        :returns: Contents requested stored within the project.
        """

        path = IMAGES / f'{name}.svg'

        if not path.exists():
            return None

        return read_text(path)


    @classmethod
    def scripts(
        cls,
        name: str,
    ) -> Optional[str]:
        """
        Return the contents requested stored within the project.

        :param name: Which file in the project will be returned.
        :returns: Contents requested stored within the project.
        """

        path = SCRIPTS / f'{name}.js'

        if not path.exists():
            return None

        return read_text(path)


    @classmethod
    def styles(
        cls,
        name: str,
    ) -> Optional[str]:
        """
        Return the contents requested stored within the project.

        :param name: Which file in the project will be returned.
        :returns: Contents requested stored within the project.
        """

        path = STYLES / f'{name}.css'

        if not path.exists():
            return None

        return read_text(path)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..funcs import funcname



def test_funcname() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert funcname() == 'test_funcname'


    class Test:

        def test1(self) -> str:
            return funcname()

        @classmethod
        def test2(cls) -> str:
            return funcname()


    method1 = Test().test1()
    method2 = Test.test2()

    assert method1 == 'Test.test1'
    assert method2 == 'Test.test2'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .config import Config
from .files import ConfigFile
from .files import ConfigFiles
from .logger import Logger
from .logger import Message
from .params import ConfigParams
from .params import LoggerParams
from .params import Params
from .paths import ConfigPath
from .paths import ConfigPaths
from .utils import config_load
from .utils import config_path
from .utils import config_paths



__all__ = [
    'Config',
    'ConfigFile',
    'ConfigFiles',
    'ConfigParams',
    'ConfigPath',
    'ConfigPaths',
    'Logger',
    'LoggerParams',
    'Message',
    'Params',
    'config_load',
    'config_path',
    'config_paths']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from re import compile
from re import match as re_match

from .common import UNITIME
from .parse import since_time



NOTATE = compile(
    r'^(\d+(s|m|h|d|w|y))*$')

STRINT = compile(
    r'^\d+$')

STRFLT = compile(
    r'^\d+\.\d+$')



def unitime(
    input: UNITIME,
) -> int:
    """
    Return the seconds in integer format for provided input.

    Example
    -------
    >>> unitime('1d')
    86400
    >>> unitime('1y')
    31536000
    >>> unitime('1w3d4h')
    878400

    :param input: Input that will be converted into seconds.
    :returns: Seconds in integer format for provided input.
    """

    if isinstance(input, str):

        if re_match(STRINT, input):
            input = int(input)

        elif re_match(STRFLT, input):
            input = int(
                input.split('.')[0])

        elif re_match(NOTATE, input):
            input = since_time(
                'now', f'+{input}')

        else:  # NOCVR
            raise ValueError('input')

    if isinstance(input, float):
        input = int(input)

    assert isinstance(input, int)

    return input



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..lists import dedup_list
from ..lists import fuzzy_list
from ..lists import inlist



def test_inlist() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    needle = 123

    haystack = [123, 456]

    assert inlist(needle, haystack)



def test_dedup_list() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    value = [1, 1, '2', 2, 3, 3]

    dedup = dedup_list(
        value,
        update=False)

    assert dedup != value
    assert dedup == [1, '2', 2, 3]

    dedup_list(value)

    assert dedup == value
    assert dedup == [1, '2', 2, 3]



def test_fuzzy_list() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    values = ['1', '2']
    patterns = ['1*']

    matched = fuzzy_list(
        values, patterns)

    assert matched == ['1']

    assert not (
        fuzzy_list('', []))



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from copy import deepcopy
from datetime import datetime
from re import match as re_match
from typing import Optional
from typing import TYPE_CHECKING

from dateutil import parser

from snaptime import snap

from .common import NUMERISH
from .common import PARSABLE
from .common import SNAPABLE
from .common import STRINGNOW
from .utils import findtz
from .utils import strptime
from .utils import utcdatetime

if TYPE_CHECKING:
    from .time import Time  # noqa: F401



def parse_time(
    source: Optional[PARSABLE] = None,
    *,
    anchor: Optional[PARSABLE] = None,
    format: Optional[str] = None,
    tzname: Optional[str] = None,
) -> datetime:
    """
    Parse provided time value with various supported formats.

    +----------------------+----------------------------+
    | *Example*            | *Description*              |
    +----------------------+----------------------------+
    | 'now'                | Uses the current time      |
    +----------------------+----------------------------+
    | None, 'None', 'null' | Uses the current time      |
    +----------------------+----------------------------+
    | 'min', float('-inf') | Uses the minimum time      |
    +----------------------+----------------------------+
    | 'max', float('inf')  | Uses the maximum time      |
    +----------------------+----------------------------+
    | object               | Provide datetime or Time   |
    +----------------------+----------------------------+
    | 2000-01-01T00:00:00Z | Uses strptime and dateutil |
    +----------------------+----------------------------+
    | 1645843626.2         | Provide Unix epoch         |
    +----------------------+----------------------------+
    | '-1h@h+5m'           | Provide snaptime syntax    |
    +----------------------+----------------------------+

    Example
    -------
    >>> parse_time('1/1/2000 12:00am')
    datetime.datetime(2000, 1, 1, 0...

    Example
    -------
    >>> parse_time(0)
    datetime.datetime(1970, 1, 1, 0...

    :param source: Time in various forms that will be parsed.
    :param anchor: Optional relative time; for snap notation.
    :param format: Optional format when source is timestamp.
    :param tzname: Name of the timezone associated to source.
        This is not relevant in timezone included in source.
    :returns: Python datetime object containing related time.
    """


    if (source is not None
            and hasattr(source, 'source')):

        source = deepcopy(
            source.source)

        assert isinstance(
            source, datetime)


    if (isinstance(source, str)
            and re_match(NUMERISH, source)):
        source = float(source)


    tzinfo = findtz(tzname)


    if isinstance(source, (int, float)):
        source = datetime.fromtimestamp(
            float(source), tz=tzinfo)


    if source in STRINGNOW:
        source = (
            utcdatetime()
            .astimezone(tzinfo))

    if source in ['max', float('inf')]:
        source = deepcopy(
            datetime.max)

    if source in ['min', float('-inf')]:
        source = deepcopy(
            datetime.min)

    if (isinstance(source, str)
            and re_match(SNAPABLE, source)):
        return shift_time(
            notate=source,
            anchor=anchor,
            tzname=tzname)

    if isinstance(source, str):
        return string_time(
            source=source,
            formats=format,
            tzname=tzname)


    if (isinstance(source, datetime)
            and not source.tzinfo):
        source = deepcopy(source)
        source = source.replace(
            tzinfo=findtz(tzname))


    if isinstance(source, datetime):
        return source

    raise ValueError('source')  # NOCVR



def shift_time(
    notate: str,
    anchor: Optional[PARSABLE] = None,
    *,
    tzname: Optional[str] = None,
) -> datetime:
    """
    Shift provided time value using snaptime relative syntax.

    Example
    -------
    >>> shift_time('-1h', anchor='1/1/2000 12:00am')
    datetime.datetime(1999, 12, 31, 23...

    Example
    -------
    >>> shift_time('-1h@d', anchor='1/1/2000 12:00am')
    datetime.datetime(1999, 12, 31, 0...

    :param notate: Syntax compatable using snaptime library.
    :param anchor: Optional relative time; for snap notation.
    :param tzname: Name of the timezone associated to anchor.
        This is not relevant in timezone included in anchor.
    :returns: Python datetime object containing related time.
    """

    anchor = parse_time(
        anchor, tzname=tzname)

    parsed = snap(anchor, notate)

    assert parsed.tzinfo

    return parse_time(
        parsed, tzname=tzname)



def string_time(
    source: str,
    *,
    formats: Optional[str | list[str]] = None,
    tzname: Optional[str] = None,
) -> datetime:
    """
    Parse provided time value with various supported formats.

    Example
    -------
    >>> string_time('2000-01-01T00:00:00Z')
    datetime.datetime(2000, 1, 1, 0...

    Example
    -------
    >>> string_time('2000/01/01')
    datetime.datetime(2000, 1, 1, 0...

    :param source: Time in various forms that will be parsed.
    :param formats: Various formats compatable with strptime.
    :param tzname: Name of the timezone associated to source.
        This is not relevant in timezone included in source.
    :returns: Python datetime object containing related time.
    """

    if formats is not None:

        with suppress(ValueError):

            return strptime(
                source, formats)

    parsed = parser.parse(source)

    if parsed.tzinfo is None:

        tzinfo = findtz(tzname)

        parsed = parsed.replace(
            tzinfo=tzinfo)

    return parse_time(
        parsed, tzname=tzname)



def since_time(
    start: PARSABLE,
    stop: Optional[PARSABLE] = None,
) -> float:
    """
    Determine the time in seconds occurring between values.

    Example
    -------
    >>> start = parse_time(0)
    >>> stop = parse_time(1)
    >>> since_time(start, stop)
    1.0

    :param start: Time in various forms that will be parsed.
    :param stop: Time in various forms that will be parsed.
    :returns: Time in seconds occurring between the values.
    """

    start = parse_time(start)
    stop = parse_time(stop)

    delta = stop - start

    since = delta.total_seconds()

    return abs(since)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from time import sleep
from typing import TYPE_CHECKING

from pytest import fixture
from pytest import mark

from ..time import Time
from ..window import Window
from ..window import window_croniter
from ..window import window_interval
from ...types import inrepr
from ...types import instr
from ...types import lattrs

if TYPE_CHECKING:
    from ..common import PARSABLE



@fixture
def window() -> Window:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    return Window(
        window='* * * * *',
        start=300,
        stop=600,
        delay=10)



def test_Window(
    window: Window,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param window: Primary class instance for window object.
    """


    attrs = lattrs(window)

    assert attrs == [
        '_Window__window',
        '_Window__start',
        '_Window__stop',
        '_Window__anchor',
        '_Window__delay',
        '_Window__wlast',
        '_Window__wnext']


    assert inrepr(
        'window.Window object',
        window)

    assert isinstance(
        hash(window), int)

    assert instr(
        'window.Window object',
        window)


    assert window.window == '* * * * *'

    assert window.start == (
        '1970-01-01T00:05:00Z')

    assert window.stop == (
        '1970-01-01T00:10:00Z')

    assert window.anchor == window.start

    assert str(window.delay) == '10.0'

    assert window.last == (
        '1970-01-01T00:04:00Z')

    assert window.next == (
        '1970-01-01T00:05:00Z')

    assert window.latest == (
        '1970-01-01T00:09:50Z')

    assert not window.walked



def test_Window_cover(
    window: Window,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param window: Primary class instance for window object.
    """


    window = Window(
        window={'minutes': 1},
        start=window.start,
        stop=window.stop)

    assert window.last == (
        '1970-01-01T00:04:00Z')

    assert window.next == (
        '1970-01-01T00:05:00Z')

    assert not window.walked

    for count in range(100):
        if window.walked:
            break
        assert window.ready()

    assert count == 5
    assert not window.ready()

    assert window.last == (
        '1970-01-01T00:09:00Z')

    assert window.next == (
        '1970-01-01T00:10:00Z')

    assert window.walked


    window = Window(
        window='* * * * *',
        start=window.start,
        stop=window.stop)

    assert window.last == (
        '1970-01-01T00:04:00Z')

    assert window.next == (
        '1970-01-01T00:05:00Z')

    assert not window.walked

    for count in range(100):
        if window.walked:
            break
        assert window.ready()

    assert count == 5
    assert not window.ready()

    assert window.last == (
        '1970-01-01T00:09:00Z')

    assert window.next == (
        '1970-01-01T00:10:00Z')

    assert window.walked


    anchor = Time('-0s@s')


    window = Window(
        window='* * * * *',
        start=anchor.shift('+5m'),
        stop=anchor.shift('+10m'))

    assert not window.ready(False)


    window = Window(
        window={'seconds': 1},
        start=anchor.shift('-5s'),
        stop=anchor.shift('+5s'))

    assert not window.walked

    for count in range(100):
        if window.walked:
            break
        if window.ready():
            continue
        sleep(0.25)

    assert 16 <= count <= 32
    assert not window.ready()

    assert window.walked



@mark.parametrize(
    'window,anchor,backward,expect',
    [('* * * * *', 0, False, (0, 60)),
     ('* * * * *', 1, False, (0, 60)),
     ('* * * * *', 0, True, (-60, 0)),
     ('* * * * *', 1, True, (-60, 0)),
     ('* * * * *', 3559, False, (3540, 3600)),
     ('* * * * *', 3659, False, (3600, 3660)),
     ('* * * * *', 3660, False, (3660, 3720)),
     ('* * * * *', 3661, False, (3660, 3720)),
     ('0 * * * *', 3559, False, (0, 3600)),
     ('0 * * * *', 3660, False, (3600, 7200)),
     ('0 * * * *', 3661, False, (3600, 7200))])
def test_window_croniter(
    window: str,
    anchor: 'PARSABLE',
    backward: bool,
    expect: tuple[int, int],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param window: Parameters for defining scheduled time.
    :param anchor: Optionally define time anchor for window.
    :param backward: Optionally operate the window backward.
    :param expect: Expected output from the testing routine.
    """

    croniter = window_croniter(
        window, anchor, backward)

    assert croniter[0] == expect[0]
    assert croniter[1] == expect[1]



@mark.parametrize(
    'window,anchor,backward,expect',
    [({'seconds': 60}, 0, False, (0, 60)),
     ({'seconds': 60}, 1, False, (1, 61)),
     ({'seconds': 60}, 0, True, (-60, 0)),
     ({'seconds': 60}, 1, True, (-59, 1)),
     ({'seconds': 60}, 3559, False, (3559, 3619)),
     ({'seconds': 60}, 3659, False, (3659, 3719)),
     ({'seconds': 60}, 3660, False, (3660, 3720)),
     ({'seconds': 60}, 3661, False, (3661, 3721)),
     ({'hours': 1}, 3559, False, (3559, 7159)),
     ({'hours': 1}, 3660, False, (3660, 7260)),
     ({'hours': 1}, 3661, False, (3661, 7261))])
def test_window_interval(
    window: dict[str, int],
    anchor: 'PARSABLE',
    backward: bool,
    expect: tuple[int, int],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param window: Parameters for defining scheduled time.
    :param anchor: Optionally define time anchor for window.
    :param backward: Optionally operate the window backward.
    :param expect: Expected output from the testing routine.
    """

    interval = window_interval(
        window, anchor, backward)

    assert interval[0] == expect[0]
    assert interval[1] == expect[1]



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any



DictStrAny = dict[str, Any]
LDictStrAny = list[DictStrAny]



NCTrue = True
NCFalse = False
NCNone = None



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pytest import fixture
from pytest import mark
from pytest import raises

from ..crypts import Crypts
from ..params import CryptParams
from ..params import CryptsParams
from ...types import DictStrAny
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def crypts() -> Crypts:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    source: DictStrAny = {
        'default': {'phrase': Crypts.keygen()},
        'secrets': {'phrase': Crypts.keygen()}}

    params = CryptsParams(
        phrases=source)

    return Crypts(params)



def test_Crypts(
    crypts: Crypts,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param crypts: Primary class instance for the encryption.
    """


    attrs = lattrs(crypts)

    assert attrs == [
        '_Crypts__params']


    assert inrepr(
        'crypts.Crypts object',
        crypts)

    assert isinstance(
        hash(crypts), int)

    assert instr(
        'crypts.Crypts object',
        crypts)


    assert crypts.params

    assert len(crypts.keygen()) == 44



@mark.parametrize(
    'value,unique',
    [('foo', 'default'),
     ('foo', 'secrets')])
def test_Crypts_iterate(
    crypts: Crypts,
    value: str,
    unique: str,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param crypts: Primary class instance for the encryption.
    :param value: String value that will returned encrypted.
    :param unique: Unique identifier of mapping passphrase.
    """

    encrypt = (
        crypts.encrypt(value, unique))

    split = encrypt.split(';')

    assert split[1] == '1.0'
    assert split[2] == unique

    decrypt = crypts.decrypt(encrypt)

    assert decrypt == value



def test_Crypts_cover(
    crypts: Crypts,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param crypts: Primary class instance for the encryption.
    """


    crypts = Crypts()


    params = CryptParams(
        phrase=Crypts.keygen())

    crypts.create('testing', params)

    crypts.delete('testing')



def test_Crypts_raises(
    crypts: Crypts,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param crypts: Primary class instance for the encryption.
    """


    _raises = raises(ValueError)

    with _raises as reason:
        crypts.encrypt('foo', 'dne')

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    with _raises as reason:
        crypts.decrypt('foo')

    _reason = str(reason.value)

    assert _reason == 'string'


    _raises = raises(ValueError)

    with _raises as reason:
        string = '$ENCRYPT;1.1;f;oo'
        crypts.decrypt(string)

    _reason = str(reason.value)

    assert _reason == 'version'


    _raises = raises(ValueError)

    params = CryptParams(phrase='foo')

    with _raises as reason:
        crypts.create('default', params)

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    with _raises as reason:
        crypts.delete('dne')

    _reason = str(reason.value)

    assert _reason == 'unique'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pytest import mark

from ..color import Color
from ...types import lattrs



def test_Color() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    color = Color('000001')


    attrs = lattrs(color)

    assert attrs == [
        '_Color__source']


    assert repr(color) == (
        "Color('#000001')")

    assert isinstance(
        hash(color), int)

    assert str(color) == '#000001'


    assert int(color) == 1
    assert float(color) == 1.0

    assert color + 1 == '000002'
    assert color - 1 == '000000'

    assert color == '000001'
    assert color != '000000'
    assert color != int  # noqa

    assert color > 0
    assert color >= 1
    assert color <= 1
    assert color < 2

    assert color > '000000'
    assert color >= '000001'
    assert color <= '000001'
    assert color < '000002'



@mark.parametrize(
    'source,expect',
    [('ff00cc', (255, 0, 204)),
     ('ffffff', (255, 255, 255)),
     ('000000', (0, 0, 0)),
     ('ff0000', (255, 0, 0)),
     ('00ff00', (0, 255, 0)),
     ('0000ff', (0, 0, 255)),
     ('808080', (128, 128, 128)),
     ('ffff00', (255, 255, 0)),
     ('00ffff', (0, 255, 255)),
     ('ff00ff', (255, 0, 255)),
     ('800080', (128, 0, 128))])
def test_Color_rgb(
    source: str,
    expect: tuple[int, ...],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param source: Source color used when converting values.
    :param expect: Expected output from the testing routine.
    """

    assert Color(source).rgb == expect
    assert Color.from_rgb(*expect) == source



@mark.parametrize(
    'source,expect',
    [('ff00cc', (52.1391, 25.6196, 59.3238)),
     ('ffffff', (95.0500, 100.0000, 108.9000)),
     ('000000', (0.0000, 0.0000, 0.0000)),
     ('ff0000', (41.2400, 21.2600, 1.9300)),
     ('00ff00', (35.7600, 71.5200, 11.9200)),
     ('0000ff', (18.0500, 7.2200, 95.0500)),
     ('808080', (20.5175, 21.5861, 23.5072)),
     ('ffff00', (77.0000, 92.7800, 13.8500)),
     ('00ffff', (53.8100, 78.7400, 106.9700)),
     ('ff00ff', (59.2900, 28.4800, 96.9800)),
     ('800080', (12.7984, 6.1477, 20.9342))])
def test_Color_xyz(
    source: str,
    expect: tuple[int, ...],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param source: Source color used when converting values.
    :param expect: Expected output from the testing routine.
    """

    assert Color(source).xyz == expect



@mark.parametrize(
    'source,expect',
    [('ff00cc', (0.3803, 0.1869)),
     ('ffffff', (0.3127, 0.3290)),
     ('000000', (0.0000, 0.0000)),
     ('ff0000', (0.6401, 0.3300)),
     ('00ff00', (0.3000, 0.6000)),
     ('0000ff', (0.1500, 0.0600)),
     ('808080', (0.3127, 0.3290)),
     ('ffff00', (0.4193, 0.5053)),
     ('00ffff', (0.2247, 0.3287)),
     ('ff00ff', (0.3209, 0.1542)),
     ('800080', (0.3209, 0.1542))])
def test_Color_xy(
    source: str,
    expect: tuple[int, ...],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param source: Source color used when converting values.
    :param expect: Expected output from the testing routine.
    """

    assert Color(source).xy == expect



@mark.parametrize(
    'source,expect',
    [('ff00cc', (312, 100, 50)),
     ('ffffff', (0, 0, 100)),
     ('000000', (0, 0, 0)),
     ('ff0000', (0, 100, 50)),
     ('00ff00', (120, 100, 50)),
     ('0000ff', (240, 100, 50)),
     ('808080', (0, 0, 50)),
     ('ffff00', (60, 100, 50)),
     ('00ffff', (180, 100, 50)),
     ('ff00ff', (300, 100, 50)),
     ('80007F', (300, 100, 25))])
def test_Color_hsl(
    source: str,
    expect: tuple[int, ...],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param source: Source color used when converting values.
    :param expect: Expected output from the testing routine.
    """

    assert Color(source).hsl == expect
    assert Color.from_hsl(*expect) == source



def test_Color_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    color1 = Color('000001')
    color2 = Color('#000001')

    assert not (color1 > None)  # type: ignore
    assert not (color1 >= None)  # type: ignore
    assert not (color1 <= None)  # type: ignore
    assert not (color1 < None)  # type: ignore

    assert color1 - color2 == 0
    assert color1 - '000001' == 0
    assert color1 + color2 == 2
    assert color1 + '000001' == 2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from datetime import datetime
from re import compile
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from .time import Time



NUMERISH = compile(
    r'^\-?\d+(\.\d+)?$')

SNAPABLE = compile(
    r'^(\-|\+)[\d\@a-z\-\+]+$')

UNITIME = Union[
    int | float | str]

STRINGNOW = {
    'None', None,
    'null', 'now'}



NUMERIC = Union[int, float]

PARSABLE = Union[
    str, NUMERIC,
    datetime, 'Time']

SCHEDULE = Union[
    str, dict[str, int]]



UNIXEPOCH = (
    '1970-01-01T00:00:00+0000')

UNIXMPOCH = (
    '1970-01-01T00:00:00.000000+0000')

UNIXSPOCH = (
    '1970-01-01T00:00:00Z')

UNIXHPOCH = (
    '01/01/1970 12:00AM UTC')



STAMP_SIMPLE = (
    '%Y-%m-%dT%H:%M:%S%z')

STAMP_SUBSEC = (
    '%Y-%m-%dT%H:%M:%S.%f%z')

STAMP_HUMAN = (
    '%m/%d/%Y %I:%M%p %Z')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Iterator

from pytest import fixture

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.remote.webdriver import WebDriver

from ..content import Content



@fixture
def content() -> Content:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    return Content()



@fixture
def driver() -> Iterator[WebDriver]:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    options = Options()

    options.add_argument('--headless')

    driver = (
        webdriver
        .Chrome(options=options))

    yield driver

    driver.quit()



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import Annotated
from typing import Any
from typing import Optional

from pydantic import Field

from .logger import LOGLEVELS
from ..crypts import CryptsParams
from ..types import BaseModel



class ConfigParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    paths: Annotated[
        Optional[list[str]],
        Field(None,
              description='Location of configuration files',
              min_length=1)]



class LoggerParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    stdo_level: Annotated[
        Optional[LOGLEVELS],
        Field(None,
              description='Minimum logging message level',
              min_length=1)]

    file_level: Annotated[
        Optional[LOGLEVELS],
        Field(None,
              description='Minimum logging message level',
              min_length=1)]

    file_path: Annotated[
        Optional[str],
        Field(None,
              description='Enable output to the log file',
              min_length=1)]


    def __init__(
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        file_path = data.get('file_path')

        if isinstance(file_path, Path):
            data['file_path'] = str(file_path)

        super().__init__(**data)



class Params(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    enconfig: Annotated[
        Optional[ConfigParams],
        Field(None,
              description='Parameters for Config instance')]

    enlogger: Annotated[
        Optional[LoggerParams],
        Field(None,
              description='Parameters for Logger instance')]

    encrypts: Annotated[
        Optional[CryptsParams],
        Field(None,
              description='Parameters for Crypts instance')]



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from pathlib import Path



SAMPLES = (
    Path(__file__).parent
    / 'samples')



_DICT1 = {
    'dict1': 'dict1',
    'str': 'd1string',
    'list': ['d1list'],
    'tuple': (1, 2),
    'dict': {'key': 'd1dict'},
    'bool': False,
    'null': None}

_DICT2 = {
    'dict2': 'dict2',
    'str': 'd2string',
    'list': ['d2list'],
    'tuple': (3, 4),
    'dict': {'key': 'd2dict'},
    'bool': True,
    'null': 'null'}

_DICT1R = deepcopy(_DICT1)
_DICT2R = deepcopy(_DICT2)

_DICT1R['recurse'] = (
    deepcopy(_DICT1))

_DICT2R['recurse'] = (
    deepcopy(_DICT2))

_DICT1R['nested'] = [
    deepcopy(_DICT1)]

_DICT2R['nested'] = [
    deepcopy(_DICT2)]



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from ..files import append_text
from ..files import read_text
from ..files import save_text



def test_readsave_text(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    content = 'pytest'

    save_text(
        f'{tmp_path}/test.txt',
        content)

    loaded = read_text(
        f'{tmp_path}/test.txt')

    assert loaded == content



def test_append_text(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    content = 'pytest'

    append_text(
        f'{tmp_path}/test.txt',
        content)

    append_text(
        f'{tmp_path}/test.txt',
        content)

    loaded = read_text(
        f'{tmp_path}/test.txt')

    assert loaded == (
        f'{content}{content}')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..duration import Duration
from ...types import inrepr
from ...types import lattrs
from ...types.strings import COMMAS



def test_Duration() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    durate = Duration(95401)


    attrs = lattrs(durate)

    assert attrs == [
        '_Duration__source',
        '_Duration__smart',
        '_Duration__groups']


    assert inrepr(
        'Duration(seconds=95401',
        durate)

    assert isinstance(
        hash(durate), int)

    assert str(durate) == '1d2h30m'


    assert int(durate) == 95401
    assert float(durate) == 95401

    assert durate + 1 == 95402
    assert durate + durate == 190802
    assert durate - 1 == 95400
    assert durate - durate == 0

    assert durate == durate
    assert durate != Duration(60)
    assert durate != 'invalid'

    assert durate > Duration(95400)
    assert durate >= Duration(95401)
    assert durate < Duration(95402)
    assert durate <= Duration(95401)


    assert durate.source == 95401

    assert durate.smart

    assert durate.groups == 7

    assert durate.units() == {
        'day': 1,
        'hour': 2,
        'minute': 30}

    assert durate.short == '1d 2h 30m'

    assert durate.compact == '1d2h30m'

    assert durate.verbose == (
        '1 day, 2 hours, 30 minutes')



def test_Duration_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    durate = Duration(
        seconds=7501,
        smart=False)

    assert durate.short == '2h 5m 1s'
    assert durate.compact == '2h5m1s'
    assert durate.verbose == (
        '2 hours, 5 minutes, 1 second')


    durate = Duration(
        groups=3,
        seconds=694800,
        smart=False)

    assert durate.short == '1w 1d 1h'
    assert durate.compact == '1w1d1h'
    assert durate.verbose == (
        '1 week, 1 day, 1 hour')


    durate = Duration(36295261)

    units = durate.units(False)
    assert units == {
        'year': 1, 'week': 3,
        'month': 1, 'day': 4,
        'hour': 2, 'minute': 1}

    units = durate.units(True)
    assert units == {
        'y': 1, 'w': 3, 'mon': 1,
        'd': 4, 'h': 2, 'm': 1}



def test_Duration_iterate() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    second = 60
    hour = second * 60
    day = hour * 24
    week = day * 7
    month = day * 30
    quarter = day * 90
    year = day * 365


    expects = {

        year: (
            '1y', '1 year'),
        year + 1: (
            '1y', '1 year'),
        year - 1: (
            '12mon4d23h59m',
            '12 months, 4 days'),

        quarter: (
            '3mon', '3 months'),
        quarter + 1: (
            '3mon', '3 months'),
        quarter - 1: (
            '2mon4w1d23h59m',
            '2 months, 4 weeks'),

        month: (
            '1mon', '1 month'),
        month + 1: (
            '1mon', '1 month'),
        month - 1: (
            '4w1d23h59m',
            '4 weeks, 1 day'),

        week: (
            '1w', '1 week'),
        week + 1: (
            '1w', '1 week'),
        week - 1: (
            '6d23h59m',
            '6 days, 23 hours'),

        day: (
            '1d', '1 day'),
        day + 1: (
            '1d', '1 day'),
        day - 1: (
            '23h59m',
            '23 hours, 59 minutes'),

        hour: (
            '1h', '1 hour'),
        hour + 1: (
            '1h', '1 hour'),
        hour - 1: (
            '59m', '59 minutes'),

        second: (
            '1m', '1 minute'),
        second + 1: (
            '1m', '1 minute'),
        second - 1: (
            '59s', 'just now')}


    items = expects.items()

    for source, expect in items:

        durate = Duration(source)

        compact, verbose = expect
        _compact = durate.compact
        _verbose = COMMAS.join(
            durate.verbose
            .split(', ')[:2])

        assert _compact == compact
        assert _verbose == verbose



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from colorsys import hls_to_rgb
from contextlib import suppress
from typing import Union

from ..types import strplwr



class Color:
    """
    Covert colors to various forms using provided hex value.

    Example
    -------
    >>> color = Color('#003333')
    >>> color.rgb
    (0, 51, 51)
    >>> color.hsl
    (180, 100, 10)
    >>> color.xyz
    (1.7814, 2.6067, 3.5412)
    >>> color.xy
    (0.2247, 0.3287)

    Example
    -------
    >>> color1 = Color('#003333')
    >>> color2 = Color('#330000')
    >>> color1 - color2
    Color('#32CCCD')
    >>> color1 + color2
    Color('#333333')

    :param source: Source color used when converting values.
    """

    __source: str


    def __init__(
        self,
        source: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        if source[:1] == '#':
            source = source[1:]

        source = strplwr(source)

        self.__source = source


    @classmethod
    def from_hsl(
        cls,
        hue: int,
        sat: int,
        lev: int,
    ) -> 'Color':
        """
        Initialize instance for class using provided parameters.
        """

        h, s, l = (  # noqa: E741
            hue / 360,
            sat / 100,
            lev / 100)

        r, g, b = hls_to_rgb(h, l, s)

        color = (
            f'{round(r * 255):02X}'
            f'{round(g * 255):02X}'
            f'{round(b * 255):02X}')

        return cls(color)


    @classmethod
    def from_rgb(
        cls,
        red: int,
        green: int,
        blue: int,
    ) -> 'Color':
        """
        Initialize instance for class using provided parameters.
        """

        value = (
            f'{red:02X}'
            f'{green:02X}'
            f'{blue:02X}')

        return cls(value)


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        color = self.__str__()

        return f"Color('{color}')"


    def __hash__(
        self,
    ) -> int:
        """
        Built-in method used when performing hashing operations.

        :returns: Integer hash value for the internal reference.
        """

        return hash(self.__source)


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        color = self.__source.upper()

        return f'#{color}'


    def __int__(
        self,
    ) -> int:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return int(self.__source, 16)


    def __float__(
        self,
    ) -> float:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return float(self.__int__())


    def __add__(
        self,
        other: Union[int, str, 'Color'],
    ) -> 'Color':
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        if isinstance(other, str):
            other = Color(other)

        source = self.__int__()
        _source = int(other)

        outcome = abs(source + _source)

        result = f'{outcome:06x}'

        return Color(result)


    def __sub__(
        self,
        other: Union[int, str, 'Color'],
    ) -> 'Color':
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        if isinstance(other, str):
            other = Color(other)

        source = self.__int__()
        _source = int(other)

        outcome = abs(source - _source)

        result = f'{outcome:06x}'

        return Color(result)


    def __eq__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        with suppress(Exception):

            if isinstance(other, int):
                other = f'{other:06x}'

            if isinstance(other, str):
                other = Color(other)

            assert hasattr(
                other, 'source')

            source = self.__source
            _source = other.source

            assert isinstance(_source, str)

            return source == _source

        return False


    def __ne__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return not self.__eq__(other)


    def __gt__(
        self,
        other: Union[int, str, 'Color'],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        with suppress(Exception):

            if isinstance(other, int):
                other = f'{other:06x}'

            if isinstance(other, str):
                other = Color(other)

            assert hasattr(
                other, 'source')

            source = self.__source
            _source = other.source

            assert isinstance(_source, str)

            return source > _source

        return False


    def __ge__(
        self,
        other: Union[int, str, 'Color'],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        with suppress(Exception):

            if isinstance(other, int):
                other = f'{other:06x}'

            if isinstance(other, str):
                other = Color(other)

            assert hasattr(
                other, 'source')

            source = self.__source
            _source = other.source

            assert isinstance(_source, str)

            return source >= _source

        return False


    def __lt__(
        self,
        other: Union[int, str, 'Color'],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        with suppress(Exception):

            if isinstance(other, int):
                other = f'{other:06x}'

            if isinstance(other, str):
                other = Color(other)

            assert hasattr(
                other, 'source')

            source = self.__source
            _source = other.source

            assert isinstance(_source, str)

            return source < _source

        return False


    def __le__(
        self,
        other: Union[int, str, 'Color'],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        with suppress(Exception):

            if isinstance(other, int):
                other = f'{other:06x}'

            if isinstance(other, str):
                other = Color(other)

            assert hasattr(
                other, 'source')

            source = self.__source
            _source = other.source

            assert isinstance(_source, str)

            return source <= _source

        return False


    @property
    def source(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source


    @property
    def rgb(
        self,
    ) -> tuple[int, int, int]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        color = self.__source

        return (
            int(color[0:2], 16),
            int(color[2:4], 16),
            int(color[4:6], 16))


    @property
    def xyz(
        self,
    ) -> tuple[float, float, float]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        _red, _grn, _blu = self.rgb

        red = _red / 255.0
        grn = _grn / 255.0
        blu = _blu / 255.0

        if red > 0.04045:
            red += 0.055
            red /= 1.055
            red **= 2.4

        else:
            red = red / 12.92

        if grn > 0.04045:
            grn += 0.055
            grn /= 1.055
            grn **= 2.4

        else:
            grn = grn / 12.92

        if blu > 0.04045:
            blu += 0.055
            blu /= 1.055
            blu **= 2.4

        else:
            blu = blu / 12.92

        red = red * 100
        grn = grn * 100
        blu = blu * 100

        x = (
            red * 0.4124
            + grn * 0.3576
            + blu * 0.1805)

        y = (
            red * 0.2126
            + grn * 0.7152
            + blu * 0.0722)

        z = (
            red * 0.0193
            + grn * 0.1192
            + blu * 0.9505)

        x = round(x, 4)
        y = round(y, 4)
        z = round(z, 4)

        return x, y, z


    @property
    def xy(
        self,
    ) -> tuple[float, float]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        x, y, z = self.xyz

        if x + y + z == 0:
            return 0, 0

        cx = x / (x + y + z)
        cy = y / (x + y + z)

        cx = round(cx, 4)
        cy = round(cy, 4)

        return cx, cy


    @property
    def hsl(
        self,
    ) -> tuple[int, int, int]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        _red, _grn, _blu = self.rgb

        red = _red / 255
        grn = _grn / 255
        blu = _blu / 255

        mam = max(red, grn, blu)
        mim = min(red, grn, blu)

        hue = sat = 0.0
        lev = (mam + mim) / 2

        if mam != mim:

            delta = mam - mim

            sat = (
                delta / (2 - mam - mim)
                if lev > 0.5
                else delta / (mam + mim))

            if mam == red:
                hue = (grn - blu) / delta
                hue += 6 if grn < blu else 0

            elif mam == grn:
                hue = (blu - red) / delta
                hue += 2

            elif mam == blu:
                hue = (red - grn) / delta
                hue += 4

            hue /= 6

        hue = int(hue * 360)
        sat = int(sat * 100)
        lev = int(lev * 100)

        return hue, sat, lev



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from threading import Lock
from typing import Annotated
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from pydantic import Field

from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

from .common import PARSABLE
from .params import WindowsParams
from .time import Time
from .window import Window
from ..types import BaseModel
from ..types import DictStrAny
from ..types import LDictStrAny

if TYPE_CHECKING:
    from .params import WindowParams



WINDOWS = dict[str, Window]



class WindowsTable(DeclarativeBase):
    """
    Schematic for the database operations using SQLAlchemy.

    .. note::
       Fields are not completely documented for this model.
    """

    group = Column(
        String,
        primary_key=True,
        nullable=False)

    unique = Column(
        String,
        primary_key=True,
        nullable=False)

    last = Column(
        String,
        nullable=False)

    next = Column(
        String,
        nullable=False)

    update = Column(
        String,
        nullable=False)



class WindowsRecord(BaseModel):
    """
    Contain the information of the record from the database.

    :param record: Record from the SQLAlchemy query routine.
    """

    group: Annotated[
        str,
        Field(...,
              description='Group the children by category',
              min_length=1)]

    unique: Annotated[
        str,
        Field(...,
              description='Unique identifier for the child',
              min_length=1)]

    last: Annotated[
        str,
        Field(...,
              description='Last schedule for the period',
              min_length=20,
              max_length=32)]

    next: Annotated[
        str,
        Field(...,
              description='Next schedule for the period',
              min_length=20,
              max_length=32)]

    update: Annotated[
        str,
        Field(...,
              description='When the record was updated',
              min_length=20,
              max_length=32)]


    def __init__(
        self,
        record: Optional[DeclarativeBase] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        fields = [
            'group',
            'unique',
            'last',
            'next',
            'update']

        params = {
            x: getattr(record, x)
            for x in fields}


        timefs = [
            'last',
            'next',
            'update']

        params |= {
            k: (Time(v).simple
                if v else None)
            for k, v in
            params.items()
            if k in timefs}


        super().__init__(**params)



class Windows:
    """
    Track windows on unique key determining when to proceed.

    .. warning::
       This class will use an in-memory database for cache,
       unless a cache file is explicity defined.

    .. testsetup::
       >>> from .params import WindowParams
       >>> from .params import WindowsParams

    Example
    -------
    >>> source = {'one': WindowParams(window=1)}
    >>> params = WindowsParams(windows=source)
    >>> windows = Windows(params, '-2s', 'now')
    >>> [windows.ready('one') for x in range(4)]
    [True, True, True, False]

    :param params: Parameters used to instantiate the class.
    :param start: Determine the start for scheduling window.
    :param stop: Determine the ending for scheduling window.
    :param store: Optional database path for keeping state.
    :param group: Optional override for default group name.
    """

    __params: 'WindowsParams'

    __store: str
    __group: str
    __table: Type[WindowsTable]
    __locker: Lock

    __sengine: Engine
    __session: (
        # pylint: disable=unsubscriptable-object
        sessionmaker[Session])

    __start: Time
    __stop: Time

    __childs: WINDOWS


    def __init__(  # noqa: CFQ002
        self,
        params: Optional['WindowsParams'] = None,
        start: PARSABLE = 'now',
        stop: PARSABLE = '3000-01-01',
        *,
        store: str = 'sqlite:///:memory:',
        table: str = 'windows',
        group: str = 'default',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        params = deepcopy(params)

        if params is None:
            params = WindowsParams()

        self.__params = params


        self.__store = store
        self.__group = group

        class SQLBase(DeclarativeBase):
            pass

        self.__table = type(
            f'encommon_windows_{table}',
            (SQLBase, WindowsTable),
            {'__tablename__': table})

        self.__locker = Lock()

        self.__build_engine()


        start = Time(start)
        stop = Time(stop)

        assert stop > start

        self.__start = start
        self.__stop = stop


        self.__childs = {}

        self.__build_objects()


    def __build_engine(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        sengine = create_engine(
            self.__store,
            pool_pre_ping=True)

        (self.__table.metadata
         .create_all(sengine))

        session = (
            sessionmaker(sengine))

        self.__sengine = sengine
        self.__session = session


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.__params

        items = (
            params.windows
            .items())

        for name, item in items:
            self.create(name, item)


    @property
    def params(
        self,
    ) -> 'WindowsParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def store(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__store


    @property
    def group(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__group


    @property
    def store_table(
        self,
    ) -> Type[WindowsTable]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__table


    @property
    def store_engine(
        self,
    ) -> Engine:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__sengine


    @property
    def store_session(
        self,
    ) -> Session:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__session()


    @property
    def start(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__start)


    @property
    def stop(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__stop)


    @property
    def children(
        self,
    ) -> dict[str, Window]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__childs)


    def save_children(
        self,
    ) -> None:
        """
        Save the child caches from the attribute into database.
        """

        sess = self.__session()
        lock = self.__locker

        table = self.__table
        group = self.__group
        childs = self.__childs


        inserts: LDictStrAny = []

        update = Time('now')


        items = childs.items()

        for unique, cache in items:

            _last = cache.last.subsec
            _next = cache.next.subsec
            _update = update.subsec

            inputs: DictStrAny = {
                'group': group,
                'unique': unique,
                'last': _last,
                'next': _next,
                'update': _update}

            inserts.append(inputs)


        with lock, sess as session:

            for insert in inserts:

                session.merge(
                    table(**insert))

            session.commit()


    def ready(
        self,
        unique: str,
        update: bool = True,
    ) -> bool:
        """
        Determine whether or not the appropriate time has passed.

        :param unique: Unique identifier for the related child.
        :param update: Determines whether or not time is updated.
        :returns: Boolean indicating whether enough time passed.
        """

        childs = self.__childs

        if unique not in childs:
            raise ValueError('unique')

        child = childs[unique]

        ready = child.ready(update)

        if ready is True:
            self.save_children()

        return ready


    def pause(
        self,
        unique: str,
        update: bool = True,
    ) -> bool:
        """
        Determine whether or not the appropriate time has passed.

        :param unique: Unique identifier for the related child.
        :param update: Determines whether or not time is updated.
        :returns: Boolean indicating whether enough time passed.
        """

        return not self.ready(
            unique, update)


    def select(
        self,
        unique: str,
    ) -> Optional[WindowsRecord]:
        """
        Return the record from within the table in the database.

        :returns: Record from within the table in the database.
        """

        sess = self.__session()
        lock = self.__locker

        table = self.__table
        model = WindowsRecord

        table = self.__table
        group = self.__group

        _unique = table.unique
        _group = table.group


        with lock, sess as session:

            query = (
                session.query(table)
                .filter(_unique == unique)
                .filter(_group == group))

            _records = list(query.all())


        records = [
            model(x)
            for x in _records]


        if len(records) == 0:
            return None

        assert len(records) == 1

        return records[0]


    def create(
        self,
        unique: str,
        params: 'WindowParams',
    ) -> Window:
        """
        Create a new window using the provided input parameters.

        :param unique: Unique identifier for the related child.
        :param params: Parameters used to instantiate the class.
        :returns: Newly constructed instance of related class.
        """

        childs = self.__childs
        _start = self.__start
        _stop = self.__stop

        if unique in childs:
            raise ValueError('unique')

        window = params.window
        start = params.start
        stop = params.stop
        anchor = params.anchor
        delay = params.delay


        select = self.select(unique)

        if select is not None:
            start = select.next
            anchor = select.next


        child_start = Time(
            start if start
            else _start)

        child_stop = Time(
            stop if stop
            else _stop)

        anchor = anchor or start


        if child_start < _start:
            child_start = _start

        if child_stop > _stop:
            child_stop = _stop


        child = Window(
            window=window,
            start=child_start,
            stop=child_stop,
            anchor=anchor,
            delay=delay)

        childs[unique] = child


        self.save_children()

        return child


    def update(
        self,
        unique: str,
        value: Optional[PARSABLE] = None,
    ) -> None:
        """
        Update the window from the provided parasable time value.

        :param unique: Unique identifier for the related child.
        :param value: Override the time updated for window value.
        """

        childs = self.__childs

        if unique not in childs:
            raise ValueError('unique')

        child = childs[unique]

        child.update(value)

        self.save_children()


    def delete(
        self,
        unique: str,
    ) -> None:
        """
        Delete the window from the internal dictionary reference.

        .. note::
           This is a graceful method, will not raise exception
           when the provided unique value does not exist.

        :param unique: Unique identifier for the related child.
        """

        sess = self.__session()
        lock = self.__locker
        childs = self.__childs

        table = self.__table
        group = self.__group

        _unique = table.unique
        _group = table.group


        with lock, sess as session:

            (session.query(table)
             .filter(_unique == unique)
             .filter(_group == group)
             .delete())

            session.commit()


        if unique in childs:
            del childs[unique]



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from datetime import datetime
from datetime import time as dtime
from datetime import timedelta
from typing import Optional

from .common import PARSABLE
from .common import STAMP_HUMAN
from .common import STAMP_SIMPLE
from .common import STAMP_SUBSEC
from .parse import parse_time
from .parse import since_time
from .utils import findtz
from .utils import strftime



class Time:
    """
    Interact with various time functions through one wrapper.

    .. testsetup::
       >>> time = Time('1/1/2000 12:00am')

    Example
    -------
    >>> time = Time('1/1/2000 12:00am')
    >>> time.stamp()
    '2000-01-01T00:00:00.000000+0000'
    >>> time.stamp('%m/%d/%Y')
    '01/01/2000'

    Example
    -------
    >>> time.epoch
    946684800.0
    >>> time.time
    datetime.time(0, 0)
    >>> time.simple
    '2000-01-01T00:00:00+0000'
    >>> time.human
    '01/01/2000 12:00AM UTC'

    Example
    -------
    >>> time.before
    Time('1999-12-31T23:59:59.999999+0000')
    >>> time.after
    Time('2000-01-01T00:00:00.000001+0000')

    Example
    -------
    >>> time.shift('-1d')
    Time('1999-12-31T00:00:00.000000+0000')

    Example
    -------
    >>> time.shifz('US/Central')
    Time('1999-12-31T18:00:00.000000-0600')

    Example
    -------
    >>> time = Time('-1s')
    >>> int(time.since)
    1

    Example
    -------
    >>> time1 = Time('1/1/2000 12:00am')
    >>> time2 = Time('1/1/2000 12:00am')
    >>> time1 - time2
    0.0
    >>> time1 + time2
    1893369600.0

    :param source: Time in various forms that will be parsed.
    :param anchor: Optional relative time; for snap notation.
    :param format: Optional format when source is timestamp.
    :param tzname: Name of the timezone associated to source.
        This is not relevant in timezone included in source.
    """

    __source: datetime


    def __init__(
        self,
        source: Optional[PARSABLE] = None,
        *,
        anchor: Optional[PARSABLE] = None,
        format: str = STAMP_SUBSEC,
        tzname: Optional[str] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        parsed = parse_time(
            source=source,
            anchor=anchor,
            format=format,
            tzname=tzname)

        self.__source = parsed


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return f"Time('{self.subsec}')"


    def __hash__(
        self,
    ) -> int:
        """
        Built-in method used when performing hashing operations.

        :returns: Integer hash value for the internal reference.
        """

        return hash(self.__source)


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return self.stamp()


    def __int__(
        self,
    ) -> int:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return int(self.epoch)


    def __float__(
        self,
    ) -> float:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return float(self.epoch)


    def __add__(
        self,
        other: PARSABLE,
    ) -> float:
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        parsed = (
            parse_time(other)
            .timestamp())

        return self.epoch + parsed


    def __sub__(
        self,
        other: PARSABLE,
    ) -> float:
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        parsed = (
            parse_time(other)
            .timestamp())

        return self.epoch - parsed


    def __eq__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        source = self.__source

        with suppress(Exception):

            parsed = parse_time(
                other)  # type: ignore

            return source == parsed

        return False


    def __ne__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return not self.__eq__(other)


    def __gt__(
        self,
        other: PARSABLE,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        parsed = parse_time(other)

        return self.__source > parsed


    def __ge__(
        self,
        other: PARSABLE,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        parsed = parse_time(other)

        return self.__source >= parsed


    def __lt__(
        self,
        other: PARSABLE,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        parsed = parse_time(other)

        return self.__source < parsed


    def __le__(
        self,
        other: PARSABLE,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        parsed = parse_time(other)

        return self.__source <= parsed


    @property
    def source(
        self,
    ) -> datetime:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source


    @property
    def epoch(
        self,
    ) -> float:
        """
        Return the seconds since the Unix epoch for the instance.

        :returns: Seconds since the Unix epoch for the instance.
        """

        source = self.__source

        return source.timestamp()


    @property
    def spoch(
        self,
    ) -> int:
        """
        Return the seconds since the Unix epoch for the instance.

        :returns: Seconds since the Unix epoch for the instance.
        """

        return int(self.epoch)


    @property
    def mpoch(
        self,
    ) -> float:
        """
        Return milliseconds since the Unix epoch for the instance.

        :returns: Seconds since the Unix epoch for the instance.
        """

        return self.epoch * 1000


    @property
    def time(
        self,
    ) -> dtime:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source.time()


    @property
    def simple(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.stamp(STAMP_SIMPLE)


    @property
    def subsec(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.stamp(STAMP_SUBSEC)


    @property
    def human(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.stamp(STAMP_HUMAN)


    @property
    def elapsed(
        self,
    ) -> float:
        """
        Determine the time in seconds occurring since instance.

        :returns: Time in seconds occurring since the instance.
        """

        return since_time(self.__source)


    @property
    def since(
        self,
    ) -> float:
        """
        Determine the time in seconds occurring since instance.

        :returns: Time in seconds occurring since the instance.
        """

        return since_time(self.__source)


    @property
    def before(
        self,
    ) -> 'Time':
        """
        Return new object containing time just before the time.

        :returns: Object containing time just before the time.
        """

        source = self.__source

        source -= timedelta(
            microseconds=1)

        return Time(source)


    @property
    def after(
        self,
    ) -> 'Time':
        """
        Return new object containing time just after the time.

        :returns: Object containing time just after the time.
        """

        source = self.__source

        source += timedelta(
            microseconds=1)

        return Time(source)


    def stamp(
        self,
        format: str = STAMP_SUBSEC,
        tzname: Optional[str] = None,
    ) -> str:
        """
        Return the timestamp using provided format for instance.

        :param format: Optional format when source is timestamp.
        :param tzname: Name of the timezone associated to source.
            This is not relevant in timezone included in source.
        :returns: Timestamp using provided format for instance.
        """

        source = self.__source

        tzinfo = findtz(tzname)

        if tzname is not None:

            source = (
                source
                .astimezone(tzinfo))

        return strftime(
            source, format)


    def shift(
        self,
        notate: str,
    ) -> 'Time':
        """
        Return the new instance of object shifted using snaptime.

        :param notate: Syntax compatable using snaptime library.
        :returns: New instance of the class using shifted time.
        """

        source = self.__source

        return Time(
            notate, anchor=source)


    def shifz(
        self,
        tzname: str,
    ) -> 'Time':
        """
        Return the new instance of object shifted using datetime.

        :param tzname: Name of the timezone associated to source.
            This is not relevant in timezone included in source.
        :returns: New instance of the class using shifted time.
        """

        source = self.__source

        tzinfo = findtz(tzname)

        if tzname is not None:

            source = (
                source
                .astimezone(tzinfo))

        return Time(
            source, tzname=tzname)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..unitime import unitime



def test_unitime() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert unitime('1s') == 1
    assert unitime('1h') == 3600
    assert unitime('1') == 1
    assert unitime(1) == 1
    assert unitime('1.0') == 1



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import PosixPath

from ..paths import resolve_path
from ..paths import resolve_paths
from ..paths import stats_path
from ... import PROJECT
from ...times import Time



def test_resolve_path() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    path = resolve_path(
        path='/foo/bar',
        replace={'bar': 'foo'})

    assert path == (
        PosixPath('/foo/foo'))



def test_resolve_paths() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    paths = resolve_paths(
        paths='/foo/bar',
        replace={'bar': 'foo'})

    assert list(paths) == [
        PosixPath('/foo/foo')]


    paths = resolve_paths(
        paths=['/foo/bar', '/bar/foo'],
        replace={'bar': 'foo'})

    assert list(paths) == [
        PosixPath('/foo/foo')]



def test_stats_path() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    stats = stats_path(
        f'{PROJECT}/utils',
        replace={PROJECT: '/'},
        ignore=[r'\S+\.pyc'])

    stat = stats['/utils/paths.py']

    assert stat.st_ctime >= (
        Time('2023-01-01').epoch)
    assert stat.st_mtime >= (
        Time('2023-01-01').epoch)

    assert stat.st_size >= 2000

    assert list(stats) == [
        '/utils/__init__.py',
        '/utils/common.py',
        '/utils/files.py',
        '/utils/match.py',
        '/utils/paths.py',
        '/utils/sample.py',
        '/utils/stdout.py',
        '/utils/test/__init__.py',
        '/utils/test/test_files.py',
        '/utils/test/test_match.py',
        '/utils/test/test_paths.py',
        '/utils/test/test_sample.py',
        '/utils/test/test_stdout.py']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ast import literal_eval as leval
from contextlib import suppress
from copy import copy
from copy import deepcopy
from re import DOTALL
from re import findall as re_findall
from re import match as re_match
from typing import Any
from typing import Callable
from typing import Optional

from jinja2 import Environment
from jinja2 import StrictUndefined

from .network import Network
from .network import insubnet_ip
from .network import isvalid_ip
from ..colors import Color
from ..crypts import Hashes
from ..times import Duration
from ..times import Time
from ..times import unitime
from ..types import DictStrAny
from ..types import dedup_list
from ..types import fuzzy_list
from ..types import hasstr
from ..types import inlist
from ..types import instr
from ..types import merge_dicts
from ..types import rplstr
from ..types import sort_dict
from ..types import strplwr
from ..utils import fuzz_match
from ..utils import rgxp_match



FILTER = Callable[..., Any]
FILTERS = dict[str, FILTER]

JINJA2 = (
    r'(\{\{.+?\}\})|'
    r'(\{\%.+?\%\})')

LITERAL = (
    r'^((\{([^\{%].+?)?\})|(\[(.+?|)\])'
    '|True|False|None|'
    r'(\-?([1-9]\d*|0)(\.\d+)?))$')



DEFAULT: FILTERS = {

    # Python builtins
    'all': all,
    'any': any,
    'copy': copy,
    'deepcopy': deepcopy,

    # encommon.times
    'Duration': Duration,
    'Time': Time,

    # encommon.colors
    'Color': Color,

    # encommon.crypts
    'Hashes': Hashes,

    # encommon.parse
    'Network': Network,
    'insubnet_ip': insubnet_ip,
    'isvalid_ip': isvalid_ip,

    # encommon.times
    'unitime': unitime,

    # encommon.types
    'strplwr': strplwr,
    'hasstr': hasstr,
    'instr': instr,
    'inlist': inlist,
    'rplstr': rplstr,
    'dedup_list': dedup_list,
    'fuzzy_list': fuzzy_list,
    'merge_dicts': merge_dicts,
    'sort_dict': sort_dict,

    # encommon.utils
    'fuzz_match': fuzz_match,
    'rgxp_match': rgxp_match}



class Jinja2:
    """
    Parse the provided input and intelligently return value.

    Example
    -------
    >>> jinja2 = Jinja2()
    >>> jinja2.parse('{{ 0 | Time }}')
    '1970-01-01T00:00:00.000000+0000'

    :param statics: Additional values available for parsing.
    :param filters: Additional filter functions for parsing.
    """

    __statics: DictStrAny
    __filters: FILTERS

    __jinjenv: Environment


    def __init__(
        self,
        statics: Optional[DictStrAny] = None,
        filters: Optional[FILTERS] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        statics = dict(statics or {})
        filters = dict(filters or {})

        items = DEFAULT.items()

        for key, filter in items:
            filters[key] = filter

        self.__statics = statics
        self.__filters = filters

        jinjenv = Environment(
            auto_reload=False,
            autoescape=False,
            cache_size=0,
            extensions=[
                'jinja2.ext.i18n',
                'jinja2.ext.loopcontrols',
                'jinja2.ext.do'],
            keep_trailing_newline=False,
            lstrip_blocks=False,
            newline_sequence='\n',
            optimized=True,
            trim_blocks=False,
            undefined=StrictUndefined)

        jinjenv.filters |= filters

        self.__jinjenv = jinjenv


    @property
    def statics(
        self,
    ) -> DictStrAny:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__statics)


    @property
    def filters(
        self,
    ) -> FILTERS:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__filters)


    @property
    def jinjenv(
        self,
    ) -> Environment:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__jinjenv


    def parser(
        self,
        value: str,
        statics: Optional[DictStrAny] = None,
    ) -> Any:
        """
        Return the provided input using the Jinja2 environment.

        :param value: Input that will be processed and returned.
        :param statics: Additional values available for parsing.
        :returns: Provided input using the Jinja2 environment.
        """

        statics = (
            dict(statics or {})
            | self.__statics)

        parser = (
            self.__jinjenv
            .from_string)

        rendered = (
            parser(value)
            .render(**statics))

        return rendered


    def parse(  # noqa: CFQ004
        self,
        value: Any,
        statics: Optional[DictStrAny] = None,
        literal: bool = True,
    ) -> Any:
        """
        Return the provided input using the Jinja2 environment.

        :param value: Input that will be processed and returned.
        :param statics: Additional values available for parsing.
        :param literal: Determine if Python objects are evaled.
        :returns: Provided input using the Jinja2 environment.
        """


        def _final(  # noqa: CFQ004
            value: Any,
        ) -> Any:

            if literal is False:
                return value

            match = re_match(
                LITERAL, str(value))

            if match is None:
                return value

            with suppress(Exception):
                return leval(value)

            return value


        def _parse(
            value: Any,
        ) -> Any:

            return self.parse(
                value,
                statics, literal)


        def _parser(
            value: Any,
        ) -> Any:

            parsed = self.parser(
                value, statics)

            return _final(parsed)


        def _found(
            value: Any,
        ) -> list[Any]:

            value = str(value)

            return re_findall(
                JINJA2, value, DOTALL)


        if not len(_found(value)):
            return _final(value)


        with suppress(Exception):
            value = _final(value)


        if isinstance(value, dict):

            value = dict(value)

            items = value.items()

            for key, _value in items:

                _value = _parse(_value)

                value[key] = _value


        elif isinstance(value, list):

            value = list(value)

            values = enumerate(value)

            for idx, _value in values:

                _value = _parse(_value)

                value[idx] = _value


        elif value is not None:
            value = _parser(value)


        return value


    def set_static(
        self,
        key: str,
        value: Optional[Any] = None,
    ) -> None:
        """
        Simply add the provided static into internal reference.

        :param key: Where item will be inserted into reference.
        :param value: Item that will be inserted into internal.
        """

        statics = self.__statics

        if value is None:

            if key in statics:
                del statics[key]

            return None

        statics[key] = value


    def set_filter(
        self,
        key: str,
        value: Optional[FILTER] = None,
    ) -> None:
        """
        Simply add the provided filter into internal reference.

        :param key: Where item will be inserted into reference.
        :param value: Item that will be inserted into internal.
        """

        jinjenv = self.__jinjenv

        filters = jinjenv.filters

        if value is None:

            if key in filters:
                del filters[key]

            return None

        filters[key] = value



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from time import sleep

from pytest import fixture
from pytest import raises

from ..params import WindowParams
from ..windows import Windows
from ...types import DictStrAny
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def windows(
    tmp_path: Path,
) -> Windows:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    model = WindowParams

    source: DictStrAny = {
        'one': model(
            window='* * * * *',
            start=300,
            stop=610,
            delay=10),
        'two': model(
            window='* * * * *',
            start=300,
            stop=620,
            delay=10)}


    store = (
        f'sqlite:///{tmp_path}'
        '/cache.db')

    windows = Windows(
        start=310,
        stop=610,
        store=store)

    table = (
        windows
        .store_table)

    session = (
        windows
        .store_session)


    record = table(
        group='default',
        unique='two',
        last='1970-01-01T00:06:00Z',
        next='1970-01-01T00:07:00Z',
        update='1970-01-01T01:00:00Z')

    session.merge(record)

    record = table(
        group='default',
        unique='tre',
        last='1970-01-01T00:06:00Z',
        next='1970-01-01T00:07:00Z',
        update='1970-01-01T01:00:00Z')

    session.merge(record)

    session.commit()


    windows.create(
        'one', source['one'])

    windows.create(
        'two', source['two'])


    return windows



def test_Windows(
    windows: 'Windows',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param windows: Primary class instance for window object.
    """


    attrs = lattrs(windows)

    assert attrs == [
        '_Windows__params',
        '_Windows__store',
        '_Windows__group',
        '_Windows__table',
        '_Windows__locker',
        '_Windows__sengine',
        '_Windows__session',
        '_Windows__start',
        '_Windows__stop',
        '_Windows__childs']


    assert inrepr(
        'windows.Windows object',
        windows)

    assert isinstance(
        hash(windows), int)

    assert instr(
        'windows.Windows object',
        windows)


    assert windows.params

    assert windows.store[:6] == 'sqlite'

    assert windows.group == 'default'

    assert windows.store_table

    assert windows.store_engine

    assert windows.store_session

    assert windows.start == (
        '1970-01-01T00:05:10Z')

    assert windows.stop == (
        '1970-01-01T00:10:10Z')

    assert len(windows.children) == 2


    window = windows.children['one']

    assert window.next == (
        '1970-01-01T00:06:00Z')

    assert window.last == (
        '1970-01-01T00:05:00Z')


    window = windows.children['two']

    assert window.next == (
        '1970-01-01T00:07:00Z')

    assert window.last == (
        '1970-01-01T00:06:00Z')



def test_Windows_cover(
    windows: Windows,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param windows: Primary class instance for window object.
    """


    assert windows.ready('two')
    assert windows.ready('two')
    assert windows.ready('two')
    assert windows.pause('two')

    windows.update('two', '+1h')

    assert not windows.ready('two')
    assert windows.pause('two')


    windows = Windows()


    params = WindowParams(
        window=1,
        start='+1s')

    windows.create('fur', params)

    assert not windows.ready('fur')
    assert windows.pause('fur')

    sleep(2)

    assert windows.ready('fur')
    assert windows.ready('fur')
    assert windows.pause('fur')

    windows.delete('fur')



def test_Windows_raises(
    windows: Windows,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param windows: Primary class instance for window object.
    """


    _raises = raises(ValueError)

    with _raises as reason:
        windows.ready('dne')

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    params = WindowParams(
        window='* * * * *')

    with _raises as reason:
        windows.create('one', params)

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    with _raises as reason:
        windows.update('dne', 'now')

    _reason = str(reason.value)

    assert _reason == 'unique'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any



def merge_dicts(
    dict1: dict[Any, Any],
    dict2: dict[Any, Any],
    force: bool | None = False,
    *,
    merge_list: bool = True,
    merge_dict: bool = True,
    paranoid: bool = False,
) -> dict[Any, Any]:
    """
    Recursively merge the contents of provided dictionaries.

    .. warning::
       This function will update the ``dict1`` reference.

    Example
    -------
    >>> dict1 = {'a': 'b', 'c': [1]}
    >>> dict2 = {'a': 'B', 'c': [2]}
    >>> merge_dicts(dict1, dict2)
    {'a': 'b', 'c': [1, 2]}
    >>> dict1
    {'a': 'b', 'c': [1, 2]}

    :param dict1: Primary dictionary which is merged into.
    :param dict2: Secondary dictionary for primary updates.
    :param force: Force overwriting concrete values in the
        primary dictionary with those from secondary. When
        ``None`` only overwrites if destination is ``None``.
    :param merge_list: Determines if merged or overwritten.
    :param merge_dict: Determines if merged or overwritten.
    :param paranoid: Perform an initial deepcopy on both of
        the provided dictionaries before performing merges.
    :returns: Provided dictionary with the other merged in.
    """

    if paranoid is True:
        dict1 = deepcopy(dict1)
        dict2 = deepcopy(dict2)


    for key, value in dict2.items():

        if key not in dict1:
            dict1[key] = value

        elif (dict1[key] is None
                and force is None):
            dict1[key] = value

        elif (isinstance(dict1[key], list)
                and isinstance(value, list)
                and merge_list is True):

            dict1[key] = (
                [] + dict1[key] + value)

        elif (isinstance(dict1[key], dict)
                and isinstance(value, dict)
                and merge_dict is True):

            merge_dicts(
                dict1[key], value, force)

        elif force is True:
            dict1[key] = value


    return dict1



def sort_dict(
    value: dict[Any, Any],
    reverse: bool = False,
) -> dict[Any, Any]:
    """
    Sort the keys within the dictionary and return new one.

    Example
    -------
    >>> foo = {'b': 'be', 'a': 'ey'}
    >>> sort_dict(foo)
    {'a': 'ey', 'b': 'be'}
    >>> sort_dict(foo, True)
    {'b': 'be', 'a': 'ey'}

    :param value: Dictionary whose keys are sorted into new.
    :param reverse: Optionally reverse the sort direction.
    :returns: New dictionary with keys sorted alphabetical.
    """

    items = sorted(
        value.items(),
        reverse=reverse)

    return dict(items)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from os import stat_result
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

from .match import rgxp_match
from ..types import rplstr
from ..types import sort_dict

if TYPE_CHECKING:
    from .common import PATHABLE
    from .common import REPLACE



STATS_PATH = dict[str, stat_result]



def resolve_path(
    path: str | Path,
    replace: Optional['REPLACE'] = None,
) -> Path:
    """
    Resolve the provided path replacing the magic keywords.

    Example
    -------
    >>> resolve_path('/foo/bar')
    PosixPath('/foo/bar')

    :param path: Complete or relative path for processing.
    :param replace: Optional values to replace in the path.
    :returns: New resolved filesystem path object instance.
    """

    path = str(path).strip()

    if replace is not None:

        items = replace.items()

        for old, new in items:

            old = str(old)
            new = str(new)

            path = rplstr(
                path, old, new)

    return Path(path).resolve()



def resolve_paths(
    paths: 'PATHABLE',
    replace: Optional['REPLACE'] = None,
) -> tuple[Path, ...]:
    """
    Resolve the provided paths replacing the magic keywords.

    .. note::
       This will remove duplicative paths from the returned.

    Example
    -------
    >>> resolve_paths(['/foo/bar'])
    (PosixPath('/foo/bar'),)

    :param paths: Complete or relative paths for processing.
    :param replace: Optional values to replace in the path.
    :returns: New resolved filesystem path object instances.
    """

    returned: list[Path] = []

    if isinstance(paths, str | Path):
        paths = [paths]

    for path in paths:

        resolved = resolve_path(
            path, replace)

        if resolved in returned:
            continue

        returned.append(resolved)

    return tuple(returned)



def stats_path(
    path: str | Path,
    replace: Optional['REPLACE'] = None,
    ignore: Optional[list[str]] = None,
) -> STATS_PATH:
    """
    Collect stats object for the complete or relative path.

    .. testsetup::
       >>> from .files import save_text
       >>> path = Path(getfixture('tmpdir'))
       >>> file = path.joinpath('hello.txt')
       >>> save_text(file, 'Hello world!')
       'Hello world!'

    Example
    -------
    >>> replace = {str(path): '/'}
    >>> stats = stats_path(path, replace)
    >>> stats['/hello.txt'].st_size
    12

    :param path: Complete or relative path for enumeration.
    :param replace: Optional values to replace in the path.
    :param ignore: Paths matching these patterns are ignored.
    :returns: Metadata for files recursively found in path.
    """

    path = Path(path).resolve()

    returned: STATS_PATH = {}


    def _ignore() -> bool:

        assert ignore is not None

        return rgxp_match(
            str(item), ignore)


    for item in path.iterdir():

        if ignore and _ignore():
            continue

        if item.is_dir():

            meta = stats_path(
                item, replace, ignore)

            returned.update(meta)

        elif item.is_file():

            key = resolve_path(
                item, replace)

            stat = item.stat()

            returned[str(key)] = stat


    return sort_dict(returned)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from glob import glob
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

from .files import ConfigFile
from .utils import config_path
from .utils import config_paths
from ..types import DictStrAny
from ..types import sort_dict

if TYPE_CHECKING:
    from ..utils.common import PATHABLE



class ConfigPath:
    """
    Contain the configuration content from filesystem path.

    :param path: Complete or relative path to configuration.
    """

    path: Path
    config: dict[str, ConfigFile]


    def __init__(
        self,
        path: str | Path,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.path = config_path(path)

        glob_path = [
            f'{self.path}/*.yml',
            f'{self.path}/**/*.yml']

        self.config = {
            str(y): ConfigFile(y)
            for x in glob_path
            for y in glob(x, recursive=True)}



class ConfigPaths:
    """
    Enumerate paths and store the contents on relative path.

    .. note::
       Class can be empty in order to play nice with parent.

    :param paths: Complete or relative path to config paths.
    :param force: Force the merge on earlier files by later.
    """

    paths: tuple[Path, ...]
    config: dict[str, ConfigPath]

    __merge: Optional[DictStrAny]


    def __init__(
        self,
        paths: 'PATHABLE',
        force: bool = False,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.paths = config_paths(paths)

        self.config = {
            str(x): ConfigPath(x)
            for x in self.paths}

        self.__merge = None


    @property
    def merge(
        self,
    ) -> DictStrAny:
        """
        Return the configuration in dictionary format for paths.

        :returns: Configuration in dictionary format for paths.
        """

        config = self.config
        merge = self.__merge

        if merge is not None:
            return deepcopy(merge)

        merge = {}


        for path in config.values():

            items = path.config.items()

            for key, file in items:

                merge[key] = file.config


        merge = sort_dict(merge)

        self.__merge = merge

        return deepcopy(merge)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..match import fuzz_match
from ..match import rgxp_match



def test_rgxp_match() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert rgxp_match('1', '1')
    assert not rgxp_match('1', ['2', '3'])
    assert rgxp_match('1', ['1', '1'])
    assert not rgxp_match(['1', '2'], ['3'])
    assert not rgxp_match(['1', '2'], ['1'])
    assert rgxp_match(['1', '2'], ['1', '2'])

    assert not rgxp_match('11', '1', True)
    assert rgxp_match('1', '1', True)



def test_fuzz_match() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert fuzz_match('1', '1')
    assert not fuzz_match('1', ['2', '3'])
    assert fuzz_match('1', ['1', '1'])
    assert not fuzz_match(['1', '2'], ['3'])
    assert not fuzz_match(['1', '2'], ['1'])
    assert fuzz_match(['1', '2'], ['1', '2'])

    assert fuzz_match('11', '1*')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from pytest import fixture

from selenium.webdriver.remote.webdriver import WebDriver

from . import _executejs
from ...types.strings import SEMPTY

if TYPE_CHECKING:
    from ..content import Content



@fixture
def prepare(
    content: 'Content',
    driver: WebDriver,
) -> tuple['Content', WebDriver]:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    driver.get(
        f"""
        data:text/html;
        charset=utf-8,
        <html>
          <head>
            <script>
        {content.scripts('default')}
        {content.scripts('helpers')}
            </script>
          </head>
          <body>
            <div>Hello</div>
            <span>World</span>
          </body>
        </html>
        """)

    return (content, driver)



def test_default_create(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        '$("<p/>")[0]'
        '.tagName;')

    assert outcome == 'P'



def test_default_styles(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        '$("body>div")'
        '.css("color", "blue")'
        '.css("color", "red");',
        wrap=False)

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.style.color;')

    assert outcome == 'red'



def test_default_addClass(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        '$("body>div")'
        '.addClass("pytest")')

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.classList;')

    assert len(outcome) == 1



def test_default_remClass(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        '$("body>div")'
        '.remClass("pytest")')

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.classList;')

    assert len(outcome) == 0



def test_default_showhide(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.style.display;')

    assert outcome == SEMPTY

    _executejs(
        driver,
        '$("body>div").hide();',
        wrap=False)

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.style.display;')

    assert outcome == 'none'

    _executejs(
        driver,
        '$("body>div")'
        '.hide().show();',
        wrap=False)

    outcome = _executejs(
        driver,
        '$("body>div")[0]'
        '.style.display;')

    assert outcome == SEMPTY



def test_default_content(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare


    _executejs(
        driver,
        '$("body>div")'
        '.html("<p>pytest</p>");')

    outcome = _executejs(
        driver,
        '$("body>div").text();')

    assert outcome == 'pytest'


    _executejs(
        driver,
        '$("body>span")'
        '.text("pytest");')

    outcome = _executejs(
        driver,
        '$("body>span").text();')

    assert outcome == 'pytest'



def test_default_append(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        'let element = $("<p/>");'
        'let target = $("body");'
        'element.text("pytest");'
        'target.append(element);',
        wrap=False)

    outcome = _executejs(
        driver,
        '$("body>p")[0]'
        '.textContent;')

    assert outcome == 'pytest'



def test_default_replace(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        'let element = $("<p/>");'
        'let target = $("body");'
        'element.text("pytest");'
        'target.replace(element);',
        wrap=False)

    outcome = _executejs(
        driver,
        '$("body>p")[0]'
        '.textContent;')

    assert outcome == 'pytest'



def test_default_attr(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        '$("body")'
        '.attr("foo", "bar")')

    outcome = _executejs(
        driver,
        'return $("body")'
        ' .attr("foo");',
        wrap=False)

    assert outcome == 'bar'



def test_default_prop(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        '$("body")'
        '.prop("foo", "bar")')

    outcome = _executejs(
        driver,
        'return $("body")'
        ' .prop("foo");',
        wrap=False)

    assert outcome == 'bar'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from re import compile
from re import match as re_match
from re import sub as re_sub
from typing import Optional
from typing import TYPE_CHECKING

from cryptography.fernet import Fernet

from ..types.strings import SEMPTY

if TYPE_CHECKING:
    from .params import CryptParams
    from .params import CryptsParams



ENCRYPT = compile(
    r'^\$ENCRYPT;1\.\d;\S+\;.+?$')



class Crypts:
    """
    Encrypt and decrypt values using passphrase dictionary.

    .. testsetup::
       >>> from .params import CryptsParams

    Example
    -------
    >>> phrase = Crypts.keygen()
    >>> source = {'default': {'phrase': phrase}}
    >>> params = CryptsParams(phrases=source)
    >>> crypts = Crypts(params)
    >>> encrypt = crypts.encrypt('example')
    >>> encrypt
    '$ENCRYPT;1.0;default;...
    >>> crypts.decrypt(encrypt)
    'example'

    :param params: Parameters used to instantiate the class.
    """

    __params: 'CryptsParams'


    def __init__(
        self,
        params: Optional['CryptsParams'] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        from .params import (
            CryptsParams)

        if params is None:
            params = CryptsParams()

        self.__params = (
            deepcopy(params))


    @property
    def params(
        self,
    ) -> 'CryptsParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    def encrypt(
        self,
        value: str,
        unique: str = 'default',
    ) -> str:
        """
        Encrypt the provided value with the relevant passphrase.

        :param value: String value that will returned encrypted.
        :param unique: Unique identifier of mapping passphrase.
        :returns: Encrypted value using the relevant passphrase.
        """

        phrases = (
            self.params
            .phrases)

        if unique not in phrases:
            raise ValueError('unique')

        phrase = (
            phrases[unique]
            .phrase)

        encoded = value.encode()

        encrypt = (
            Fernet(phrase)
            .encrypt(encoded)
            .decode())

        return (
            '$ENCRYPT;1.0;'
            f'{unique};{encrypt}')


    def decrypt(
        self,
        value: str,
    ) -> str:
        """
        Decrypt the provided value with the relevant passphrase.

        :param value: String value that will returned decrypted.
        :returns: Decrypted value using the relevant passphrase.
        """

        phrases = (
            self.params
            .phrases)

        value = crypt_clean(value)

        match = re_match(
            ENCRYPT, value)

        if match is None:
            raise ValueError('string')

        version, unique, value = (
            value.split(';')[1:])

        if version != '1.0':
            raise ValueError('version')

        phrase = (
            phrases[unique]
            .phrase)

        encoded = value.encode()

        return (
            Fernet(phrase)
            .decrypt(encoded)
            .decode())


    def create(
        self,
        unique: str,
        params: 'CryptParams',
    ) -> None:
        """
        Create a new phrase using the provided input parameters.

        :param unique: Unique identifier of mapping passphrase.
        :param params: Parameters used to instantiate the class.
        """

        phrases = (
            self.params
            .phrases)

        if unique in phrases:
            raise ValueError('unique')

        phrases[unique] = params


    def delete(
        self,
        unique: str,
    ) -> None:
        """
        Delete the phrase from the internal dictionary reference.

        :param unique: Unique identifier of mapping passphrase.
        """

        phrases = (
            self.params
            .phrases)

        if unique not in phrases:
            raise ValueError('unique')

        del phrases[unique]


    @classmethod
    def keygen(
        cls: object,
    ) -> str:
        """
        Return new randomly generated Fernet key for passphrase.

        :returns: Randomly generated Fernet key for passphrase.
        """

        key = Fernet.generate_key()

        return key.decode()



def crypt_clean(
    value: str,
) -> str:
    """
    Return the parsed and normalized encrypted string value.

    :param value: String value that will returned decrypted.
    :returns: Parsed and normalized encrypted string value.
    """

    return re_sub(r'[\n\s]', SEMPTY, value)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from pytest import fixture

from .config import Config
from .config.test import SAMPLES
from .utils import save_text



def config_factory(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    content = (
        f"""

        enconfig:
          paths:
            - '{SAMPLES}/stark'
            - '{SAMPLES}/wayne'

        enlogger:
          stdo_level: debug

        encrypts:
          phrases:
            default:
              phrase: phrase

        """)

    config_path = (
        tmp_path / 'config.yml')

    config_log = (
        tmp_path / 'config.log')

    save_text(
        config_path, content)

    logger = {
        'file_path': config_log,
        'file_level': 'info'}

    cargs = {'enlogger': logger}
    sargs = {'cus/tom': 'fart'}

    return Config(
        config_path,
        cargs=cargs,
        sargs=sargs)



@fixture
def config_path(
    tmp_path: Path,
) -> Path:
    """
    Construct the directory and files needed for the tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: New resolved filesystem path object instance.
    """

    config_factory(tmp_path)

    return tmp_path.resolve()



@fixture
def config(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    return config_factory(tmp_path)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from typing import Any
from typing import Optional

from netaddr import IPAddress
from netaddr import IPNetwork



class Network:
    """
    Convert the network into the various supported formats.

    :param source: Network IPv4 or IPv6 network or address.
    """

    __source: IPNetwork


    def __init__(
        self,
        source: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        network = IPNetwork(source)

        self.__source = network


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return f"Network('{self.address_cidr}')"


    def __hash__(
        self,
    ) -> int:
        """
        Built-in method used when performing hashing operations.

        :returns: Integer hash value for the internal reference.
        """

        return hash(self.__source)


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return self.address_cidr


    def __eq__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        source = self.__source

        with suppress(Exception):

            other = IPNetwork(str(other))

            return source == other

        return False


    def __ne__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return not self.__eq__(other)


    @property
    def source(
        self,
    ) -> IPNetwork:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source


    @property
    def version(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source.version


    @property
    def cidr(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        return source.prefixlen


    @property
    def address(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        return str(source.ip)


    @property
    def address_cidr(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        address = self.address

        return f'{address}/{self.cidr}'


    @property
    def address_host(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        address = self.address

        prefix = (
            128
            if self.version == 6
            else 32)

        return f'{address}/{prefix}'


    @property
    def network(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        return str(source.network)


    @property
    def network_cidr(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        network = self.network

        return f'{network}/{self.cidr}'


    @property
    def broadcast(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        address = source.broadcast

        if address is None:
            return None

        return str(address)


    @property
    def padded(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.version != 4:
            raise ValueError('version')

        address = self.address

        octets = address.split('.')

        pads = [
            x.zfill(3)
            for x in octets]

        return '.'.join(pads)


    @property
    def reverse(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.version != 4:
            raise ValueError('version')


        address = self.address

        octets = address.split('.')

        reverse = list(reversed(octets))

        return '.'.join(reverse)


    @property
    def hwaddr(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.version != 4:
            raise ValueError('version')

        padded = self.padded

        nodots = (
            padded
            .replace('.', ''))

        ranged = range(
            0, len(nodots), 2)

        pairs = [
            nodots[x:x + 2]
            for x in ranged]

        return '-'.join(pairs)


    @property
    def netmask(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        return str(source.netmask)



    @property
    def ispublic(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether instance is state.
        """

        return (
            self.source.ip
            .is_global())


    @property
    def isprivate(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether instance is state.
        """

        return not self.ispublic


    @property
    def islinklocal(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether instance is state.
        """

        return (
            self.source
            .is_link_local())


    @property
    def islocalhost(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether instance is state.
        """

        return (
            self.source
            .is_loopback())



def insubnet_ip(
    address: str,
    networks: str | list[str] | tuple[str, ...],
) -> bool:
    """
    Return the boolean indicating address is in the network.

    :param address: Provided address to find in the network.
    :param network: Networks which values can be within any.
    :returns: Boolean indicating address is in the network.
    """

    if isinstance(networks, str):
        networks = [networks]

    networks = list(networks)

    if not isvalid_ip(address):
        raise ValueError('address')

    naddr = Network(address)

    if (naddr.version == 4
            and naddr.cidr == 32):
        address = naddr.address

    if (naddr.version == 6
            and naddr.cidr == 128):
        address = naddr.address

    parsed = IPAddress(address)

    return any(
        parsed in IPNetwork(x)
        for x in networks)



def isvalid_ip(
    value: Any,  # noqa: ANN401
) -> bool:
    """
    Return the boolean indicating whether the value is valid.

    :param value: Value that will be validated as an address.
    :returns: Boolean indicating whether the value is valid.
    """

    value = str(value)

    try:
        Network(value)
        return True

    except Exception:
        return False



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional



class EmptyType:
    """
    Useful for representing empty or absent value with typing.

    .. testsetup::
       >>> from copy import deepcopy

    Example
    -------
    >>> 'foo' if Empty else 'bar'
    'bar'

    Example
    -------
    >>> deepcopy(Empty) is Empty
    True
    """

    __empty: Optional['EmptyType'] = None


    def __init__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__empty = None


    def __new__(
        cls,
    ) -> 'EmptyType':
        """
        Built-in method called when creating new class instance.

        :returns: Same instance of class that first instantiated.
        """

        empty = cls.__empty

        if empty is not None:
            return empty

        cls.__empty = (
            object.__new__(cls))

        return cls.__empty


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return 'Empty'


    def __hash__(
        self,
    ) -> int:
        """
        Built-in method used when performing hashing operations.

        :returns: Integer hash value for the internal reference.
        """

        return hash(EmptyType)


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return 'Empty'


    def __bool__(
        self,
    ) -> bool:
        """
        Built-in method representing boolean value for instance.

        :returns: Boolean indicating outcome from the operation.
        """

        return False


    def __eq__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return isinstance(other, type(self))


    def __ne__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return not self.__eq__(other)



Empty = EmptyType()



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any

from pytest import fixture
from pytest import mark
from pytest import raises

from ..jinja2 import Jinja2
from ... import PROJECT
from ...times.common import UNIXMPOCH
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def jinja2() -> Jinja2:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """


    def strval(
        # NOCVR,
        input: str,
    ) -> str:
        return str(input)


    return Jinja2(
        {'PROJECT': PROJECT},
        {'strval': strval})



def test_Jinja2(
    jinja2: Jinja2,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param jinja2: Parsing class for the Jinja2 templating.
    """


    attrs = lattrs(jinja2)

    assert attrs == [
        '_Jinja2__statics',
        '_Jinja2__filters',
        '_Jinja2__jinjenv']


    assert inrepr(
        'jinja2.Jinja2 object',
        jinja2)

    assert isinstance(
        hash(jinja2), int)

    assert instr(
        'jinja2.Jinja2 object',
        jinja2)


    assert jinja2.statics

    assert jinja2.filters

    assert jinja2.jinjenv

    assert jinja2.parse('1') == 1



def test_Jinja2_recurse(
    jinja2: Jinja2,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param jinja2: Parsing class for the Jinja2 templating.
    """

    parse = jinja2.parse

    source = {
        'dict': {'key': '{{ "val" }}'},
        'list': [1, '{{ 2 }}', 3, '4'],
        'deep': '{{ {"foo": "bar"} }}'}

    parsed = parse(source)

    assert parsed == {
        'dict': {'key': 'val'},
        'list': [1, 2, 3, 4],
        'deep': {'foo': 'bar'}}

    _parsed = parse(f'{source}')

    assert _parsed == parsed



@mark.parametrize(
    'value,expect',
    [('{{ none }}', None),
     ('{{ 123 }}', 123),
     ('{{ 1.2 }}', 1.2),
     ('{{ -1.2 }}', -1.2)])
def test_Jinja2_literal(
    jinja2: Jinja2,
    value: Any,  # noqa: ANN401
    expect: Any,  # noqa: ANN401
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param jinja2: Parsing class for the Jinja2 templating.
    :param value: Input that will be processed and returned.
    :param expect: Expected output from the testing routine.
    """

    parse = jinja2.parse

    parsed = parse(value)

    parsed = parse(
        value,
        literal=False)

    assert parsed == str(expect)

    parsed = parse(
        value,
        literal=True)

    assert parsed == expect



@mark.parametrize(
    'value,expect',
    [('foo', 'foo'),
     ('1.0.0', '1.0.0'),
     ('1', 1),
     ('1.0', 1.0),
     ({'a': 'b'}, {'a': 'b'}),
     ("{'a': 'b'}", {'a': 'b'}),
     ('{{ 1 }}', 1),
     ('{{ "1" | float }}', 1.0),
     ('{{ "1" | int }}', 1),
     ('{{ 0 | Time }}', UNIXMPOCH),
     (100000, 100000),
     ('100000', 100000),
     (10.001, 10.001),
     ('10.001', 10.001),
     ([1, 2], [1, 2]),
     ('[1, 2]', [1, 2]),
     ('01', '01'),
     ('{{ "01" }}', '01'),
     ('-01', '-01'),
     ('{{ "-01" }}', '-01')])
def test_Jinja2_parse(
    jinja2: Jinja2,
    value: Any,  # noqa: ANN401
    expect: Any,  # noqa: ANN401
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param jinja2: Parsing class for the Jinja2 templating.
    :param value: Input that will be processed and returned.
    :param expect: Expected output from the testing routine.
    """

    parse = jinja2.parse

    parsed = parse(value)

    assert parsed == expect



def test_Jinja2_cover(
    jinja2: Jinja2,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param jinja2: Parsing class for the Jinja2 templating.
    """

    parse = jinja2.parse


    jinja2.set_static(
        'key', 'value')

    parsed = parse('{{ key }}')

    assert parsed == 'value'

    jinja2.set_static('key')

    with raises(Exception):
        parse('{{ key }}')


    jinja2.set_filter(
        'float', float)

    parsed = parse(
        '{{ 1 | float }}')

    assert parsed == 1.0

    jinja2.set_filter('float')

    with raises(Exception):
        parse('{{ 1 | float }}')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import is_dataclass
from re import compile
from re import sub as re_sub
from sys import stdout
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .common import JOINABLE
from ..times import Duration
from ..times import Time
from ..types import BaseModel
from ..types import Empty
from ..types import clsname
from ..types.strings import COMMAD
from ..types.strings import NEWLINE
from ..types.strings import SEMPTY

if TYPE_CHECKING:
    from _typeshed import DataclassInstance



ANSICODE = compile(
    r'\x1b\[[^A-Za-z]*[A-Za-z]')

ANSIARRAL = Union[
    list[Any],
    tuple[Any, ...],
    set[Any]]

ANSIARRAD = Union[
    dict[Any, Any],
    BaseModel,
    'DataclassInstance']

ANSIARRAY = Union[
    ANSIARRAL,
    ANSIARRAD]



_REPEAT = (list, tuple, dict)



@dataclass(frozen=True)
class ArrayColors:
    """
    Colors used to colorize the provided array like object.
    """

    label: int = 37
    key: int = 97

    colon: int = 37
    hyphen: int = 37

    bool: int = 93
    none: int = 33
    str: int = 92
    num: int = 93

    times: int = 96
    empty: int = 36
    other: int = 91



def print_ansi(
    string: str = SEMPTY,
    method: Literal['stdout', 'print'] = 'stdout',
    output: bool = True,
) -> str:
    """
    Print the ANSI colorized string to the standard output.

    Example
    -------
    >>> print_ansi('<c91>ERROR<c0>')
    '\\x1b[0;91mERROR\\x1b[0;0m'

    :param string: String processed using inline directives.
    :param method: Which method for standard output is used.
    :param output: Whether or not hte output should be print.
    :returns: ANSI colorized string using inline directives.
    """  # noqa: D301 LIT102

    string = make_ansi(string)

    if output is True:
        if method == 'stdout':
            stdout.write(f'{string}\n')
        else:
            print(string)  # noqa: T201

    return string



def make_ansi(
    string: str,
) -> str:
    """
    Parse the string and replace directives with ANSI codes.

    Example
    -------
    >>> make_ansi('<c91>ERROR<c0>')
    '\\x1b[0;91mERROR\\x1b[0;0m'

    :param string: String containing directives to replace.
    :returns: Provided string with the directives replaced.
    """  # noqa: D301 LIT102

    pattern = r'\<c([\d\;]+)\>'
    replace = r'\033[0;\1m'

    return re_sub(pattern, replace, string)



def kvpair_ansi(
    key: str,
    value: Any,  # noqa: ANN401
) -> str:
    """
    Process and colorize keys and values for standard output.

    Example
    -------
    >>> kvpair_ansi('k', 'v')
    '\\x1b[0;90mk\\x1b[0;37m="\\x1b[0;0m...

    :param key: String value to use for the key name portion.
    :param value: String value to use for the value portion.
    :returns: ANSI colorized string using inline directives.
    """  # noqa: D301 LIT102

    if isinstance(value, JOINABLE):
        value = COMMAD.join([
            str(x) for x in value])

    elif not isinstance(value, str):
        value = str(value)

    return make_ansi(
        f'<c90>{key}<c37>="<c0>'
        f'{value}<c37>"<c0>')



def strip_ansi(
    string: str,
) -> str:
    """
    Return the provided string with the ANSI codes removed.

    Example
    -------
    >>> strip_ansi('\\x1b[0;91mERROR\\x1b[0;0m')
    'ERROR'

    :param string: String which contains ANSI codes to strip.
    :returns: Provided string with the ANSI codes removed.
    """  # noqa: D301 LIT102

    return re_sub(ANSICODE, SEMPTY, string)



def array_ansi(  # noqa: CFQ001, CFQ004
    source: ANSIARRAY,
    *,
    indent: int = 0,
    colors: ArrayColors = ArrayColors(),
) -> str:
    """
    Print the ANSI colorized iterable to the standard output.

    .. note::
       This massive function should be refactored, possibly
       into a class with methods where there are functions.

    Example
    -------
    >>> array_ansi({'foo': 'bar'})
    "\\x1b[0;97mfoo\\x1b[0;37m:\\x1b[0;0m...

    :param source: Value in supported and iterable formats.
    :param indent: How many levels for initial indentation.
    :param colors: Determine colors used with different types.
    :returns: ANSI colorized string using inline directives.
    """  # noqa: D301 LIT102

    output: list[str] = []

    repeat = f'<c{colors.other}>REPEAT<c0>'

    space: str = ' '


    def _append(
        prefix: str,
        value: Any,  # noqa: ANN401
        indent: int,
        refers: set[int],
    ) -> None:

        if id(value) in refers:
            return output.append(
                f'{prefix} {repeat}')


        if isinstance(value, _REPEAT):
            refers.add(id(value))


        typing = {
            'list': list,
            'tuple': tuple,
            'dict': dict,
            'BaseModel': BaseModel,
            'frozenset': frozenset,
            'set': set}

        items = typing.items()

        for name, typed in items:

            if isinstance(value, BaseModel):
                name = clsname(value)

            elif is_dataclass(value):
                name = clsname(value)

            elif not isinstance(value, typed):
                continue

            output.append(
                f'{prefix} '
                f'<c{colors.label}>'
                f'{name}<c0>')

            return _process(
                source=value,
                indent=indent + 2,
                refers=refers)


        value = _concrete(value)

        output.append(
            f'{prefix} {value}')


    def _concrete(
        source: Any,  # noqa: ANN401
    ) -> str:

        color = colors.other

        if isinstance(source, bool):
            color = colors.bool

        if isinstance(source, int | float):
            color = colors.num

        elif isinstance(source, str):
            color = colors.str

        elif source is None:
            color = colors.none

        elif isinstance(source, Duration):
            color = colors.times

        elif isinstance(source, Time):
            color = colors.times

        elif source is Empty:
            color = colors.empty

        string = f'<c{color}>{source}<c0>'

        if isinstance(source, str):
            string = f"'{string}'"

        return string


    def _dict(
        source: ANSIARRAD,
        indent: int,
        refers: Optional[set[int]] = None,
    ) -> None:

        if isinstance(source, BaseModel):
            source = source.endumped

        if is_dataclass(source):
            source = asdict(source)

        assert isinstance(source, dict)

        items = source.items()

        for key, value in items:

            prefix = (
                f'{space * indent}'
                f'<c{colors.key}>{key}'
                f'<c{colors.colon}>:<c0>')

            _append(
                prefix, value, indent,
                refers=refers or set())


    def _list(
        source: ANSIARRAL,
        indent: int,
        refers: Optional[set[int]] = None,
    ) -> None:

        assert isinstance(
            source, (list, tuple, set))

        for value in source:

            prefix = (
                f'{space * indent}'
                f'<c{colors.hyphen}>-<c0>')

            _append(
                prefix, value, indent,
                refers=refers or set())


    def _process(
        source: ANSIARRAY,
        **kwargs: Any,
    ) -> None:

        if isinstance(source, dict):
            return _dict(source, **kwargs)

        if isinstance(source, BaseModel):
            return _dict(source, **kwargs)

        if is_dataclass(source):
            return _dict(source, **kwargs)

        assert isinstance(
            source,
            set | list | tuple)

        _list(source, **kwargs)


    if is_dataclass(source):

        assert not isinstance(
            source, type)

        source = asdict(source)


    if isinstance(source, dict):
        _dict(source, indent)

    elif isinstance(source, BaseModel):
        _dict(source, indent)

    elif isinstance(source, list):
        _list(source, indent)

    elif isinstance(source, tuple):
        _list(source, indent)

    elif isinstance(source, set):
        _list(source, indent)


    _output = [
        make_ansi(x) for x in output]

    return NEWLINE.join(_output)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from selenium.webdriver.remote.webdriver import WebDriver

from . import _executejs

if TYPE_CHECKING:
    from ..content import Content



def test_moderate(
    content: 'Content',
    driver: WebDriver,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param content: Primary class instance for the content.
    :param driver: Selenium driver provided by the library.
    """


    driver.get(
        f"""
        data:text/html;
        charset=utf-8,
        <html>
          <head>
            <script>
        {content.scripts('default')}
        {content.scripts('helpers')}
        {content.scripts('image')}
        {content.scripts('moderate')}
            </script>
          </head>
          <body></body>
        </html>
        """)


    outcome = _executejs(
        driver,
        'let element ='
        ' moderate(...arguments);'
        'return'
        ' element[0].outerHTML;',
        'This is a Label',
        'success',
        'some_kind_short',
        wrap=False)


    assert outcome == (
        '<div class="encommon_moderate">'
        '<div class="_icon">'
        '<div class="encommon_svgicon"'
        ' data-image="success"></div>'
        '</div>'
        '<div class="_value">'
        '<div class="_label">'
        'This is a Label</div>'
        '<div class="_small">'
        'some_kind_short</div>'
        '</div>'
        '</div>')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from pytest import fixture

from selenium.webdriver.remote.webdriver import WebDriver

from . import _executejs

if TYPE_CHECKING:
    from ..content import Content



@fixture
def prepare(
    content: 'Content',
    driver: WebDriver,
) -> tuple['Content', WebDriver]:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    driver.get(
        f"""
        data:text/html;
        charset=utf-8,
        <html>
          <head>
            <script>
        {content.scripts('default')}
        {content.scripts('helpers')}
            </script>
          </head>
          <body>
            <div>Hello</div>
            <span>World</span>
          </body>
        </html>
        """)

    return (content, driver)



def test_helpers_assert(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'assert(1 == 1)')

    assert outcome is True



def test_helpers_whenready(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    _executejs(
        driver,
        'whenready('
        'console.log)')



def test_helpers_isnull(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isnull(undefined)'
        ' && isnull(null)')

    assert outcome is True



def test_helpers_isempty(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isempty({})'
        ' && isempty("")'
        ' && isempty([])')

    assert outcome is True



def test_helpers_isbool(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isbool(true)')

    assert outcome is True



def test_helpers_isstr(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isstr("string")')

    assert outcome is True

    outcome = _executejs(
        driver,
        'isstr(69429)')

    assert outcome is False



def test_helpers_isnum(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isnum(123)')

    assert outcome is True

    outcome = _executejs(
        driver,
        'isnum(1.23)')

    assert outcome is True

    outcome = _executejs(
        driver,
        'isnum("69429")')

    assert outcome is False



def test_helpers_isquery(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        ('let element = $("<p/>");'
         'return isquery(element);'),
        wrap=False)

    assert outcome is True



def test_helpers_isnode(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        ('let element = $("<p/>");'
         'return isnode(element[0]);'),
        wrap=False)

    assert outcome is True



def test_helpers_istime(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'istime("2024-04-20")')

    assert outcome is True



def test_helpers_islist(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'islist([69429])')

    assert outcome is True



def test_helpers_isdict(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isdict({"foo": "bar"})')

    assert outcome is True



def test_helpers_istrue(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'istrue(true)')

    assert outcome is True



def test_helpers_isfalse(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'isfalse(false)')

    assert outcome is True



def test_helpers_loads(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    outcome = _executejs(
        driver,
        'loads(\'{"foo":"bar"}\')')

    assert outcome['foo'] == 'bar'



def test_helpers_dumps(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare

    dumped = '{"foo":"bar"}'

    outcome = _executejs(
        driver,
        'dumps({"foo": "bar"})')

    assert outcome == dumped



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import Union



JOINABLE = (list, tuple, set)



PATHABLE = Union[
    str, Path,
    list[str | Path],
    tuple[str | Path],
    set[str]]



REPLACE = Union[
    dict[str, str],
    dict[str, str | Path],
    dict[str, Path],
    dict[str, str],
    dict[str | Path, str],
    dict[Path, str]]



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy

from . import _DICT1
from . import _DICT1R
from . import _DICT2
from . import _DICT2R
from ..dicts import merge_dicts
from ..dicts import sort_dict



def test_merge_dicts() -> None:  # noqa: CFQ001
    """
    Perform various tests associated with relevant routines.
    """

    dict1 = deepcopy(_DICT1R)
    dict2 = deepcopy(_DICT2R)


    source = deepcopy(dict1)
    update = deepcopy(dict2)

    merge_dicts(source, update)

    assert source == {
        'dict1': 'dict1',
        'dict2': 'dict2',
        'str': 'd1string',
        'list': ['d1list', 'd2list'],
        'dict': {'key': 'd1dict'},
        'tuple': (1, 2),
        'bool': False,
        'null': None,
        'nested': [_DICT1, _DICT2],
        'recurse': {
            'dict1': 'dict1',
            'dict2': 'dict2',
            'str': 'd1string',
            'list': ['d1list', 'd2list'],
            'dict': {'key': 'd1dict'},
            'tuple': (1, 2),
            'bool': False,
            'null': None}}


    source = deepcopy(dict1)
    update = deepcopy(dict2)

    merge_dicts(source, update, True)

    assert source == {
        'dict1': 'dict1',
        'dict2': 'dict2',
        'str': 'd2string',
        'list': ['d1list', 'd2list'],
        'dict': {'key': 'd2dict'},
        'tuple': (3, 4),
        'bool': True,
        'null': 'null',
        'nested': [_DICT1, _DICT2],
        'recurse': {
            'dict1': 'dict1',
            'dict2': 'dict2',
            'str': 'd2string',
            'list': ['d1list', 'd2list'],
            'dict': {'key': 'd2dict'},
            'tuple': (3, 4),
            'bool': True,
            'null': 'null'}}


    source = deepcopy(dict1)
    update = deepcopy(dict2)

    merge_dicts(source, update, None)

    assert source == {
        'dict1': 'dict1',
        'dict2': 'dict2',
        'str': 'd1string',
        'list': ['d1list', 'd2list'],
        'dict': {'key': 'd1dict'},
        'tuple': (1, 2),
        'bool': False,
        'null': 'null',
        'nested': [_DICT1, _DICT2],
        'recurse': {
            'dict1': 'dict1',
            'dict2': 'dict2',
            'str': 'd1string',
            'list': ['d1list', 'd2list'],
            'dict': {'key': 'd1dict'},
            'tuple': (1, 2),
            'bool': False,
            'null': 'null'}}


    source = deepcopy(dict1)
    update = deepcopy(dict2)

    merge_dicts(
        source, update,
        merge_list=False,
        merge_dict=False)

    assert source == {
        'dict1': 'dict1',
        'dict2': 'dict2',
        'str': 'd1string',
        'list': ['d1list'],
        'dict': {'key': 'd1dict'},
        'tuple': (1, 2),
        'bool': False,
        'null': None,
        'nested': [_DICT1],
        'recurse': {
            'dict1': 'dict1',
            'str': 'd1string',
            'list': ['d1list'],
            'dict': {'key': 'd1dict'},
            'tuple': (1, 2),
            'bool': False,
            'null': None}}


    _source = merge_dicts(
        source, update,
        paranoid=True)

    assert source is not _source



def test_sort_dict() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert sort_dict(_DICT1) == {
        'bool': False,
        'dict': {'key': 'd1dict'},
        'tuple': (1, 2),
        'dict1': 'dict1',
        'list': ['d1list'],
        'str': 'd1string',
        'null': None}



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from time import sleep

from pytest import fixture

from ..time import Time
from ..timer import Timer
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def timer() -> Timer:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    return Timer(1)



def test_Timer(
    timer: Timer,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param timer: Primary class instance for timer object.
    """


    attrs = lattrs(timer)

    assert attrs == [
        '_Timer__timer',
        '_Timer__time']


    assert inrepr(
        'timer.Timer object',
        timer)

    assert isinstance(
        hash(timer), int)

    assert instr(
        'timer.Timer object',
        timer)


    assert timer.timer == 1

    assert timer.time >= Time('-1s')

    assert timer.since <= 1

    assert timer.remains <= 1

    assert not timer.ready()
    assert timer.pause()


    sleep(1)

    assert timer.since > 1

    assert timer.remains == 0

    assert timer.ready()
    assert timer.pause()

    assert not timer.ready()
    assert timer.pause()

    timer.update('1980-01-01')

    assert timer.ready()
    assert timer.pause()

    assert not timer.ready()
    assert timer.pause()
